"""Helper methods for authentication and authorization."""

from argparse import ONE_OR_MORE
import functools
import logging
from datetime import datetime

import requests
from authlib.common.errors import AuthlibBaseError
from authlib.jose import jwt
from fastapi import HTTPException, Depends
from fastapi.security import OAuth2PasswordBearer
from starlette.status import HTTP_401_UNAUTHORIZED

from amcat4.models import FieldSpec
from amcat4.config import get_settings, AuthOptions
from amcat4.index import ADMIN_USER, GUEST_USER, Role, get_role, get_global_role
from amcat4.fields import get_fields

oauth2_scheme = OAuth2PasswordBearer(tokenUrl="/auth/token", auto_error=False)


class InvalidToken(ValueError):
    pass


@functools.lru_cache()
def get_middlecat_config(middlecat_url) -> dict:
    r = requests.get(f"{middlecat_url}/api/configuration")
    r.raise_for_status()
    return r.json()


def verify_token(token: str) -> dict:
    """
    Verifies the given token and returns the payload

    raises a InvalidToken exception if the token could not be validated
    """
    payload = decode_middlecat_token(token)
    if missing := {"email", "resource", "exp"} - set(payload.keys()):
        raise InvalidToken(f"Invalid token, missing keys {missing}")
    now = int(datetime.now().timestamp())
    if payload["exp"] < now:
        raise InvalidToken("Token expired")
    if payload["resource"] != get_settings().host:
        raise InvalidToken(f"Wrong host! {payload['resource']} != {get_settings().host}")
    return payload


def decode_middlecat_token(token: str) -> dict:
    """
    Verifies a midddlecat token
    """
    url = get_settings().middlecat_url
    if not url:
        raise InvalidToken("No middlecat defined, cannot decrypt middlecat token")
    public_key = get_middlecat_config(url)["public_key"]
    try:
        return jwt.decode(token, public_key)
    except AuthlibBaseError as e:
        raise InvalidToken(e)


def check_global_role(user: str, required_role: Role, raise_error=True):
    """
    Check if the given user has at least the required role
    :param user: The email address of the authenticated user
    :param required_role: The minimum global role of the user
    :param raise_error: If true, raise an error when not authorized, otherwise return False
                        (will always raise an error if user is not authenticated)
    """
    if not user:
        raise HTTPException(status_code=401, detail="No authenticated user")
    try:
        global_role = get_global_role(user)
    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Error on retrieving user {user}: {e}")
    if global_role and global_role >= required_role:
        return global_role
    if raise_error:
        raise HTTPException(
            status_code=401,
            detail=f"User {user} does not have global " f"{required_role.name.title()} permissions on this instance",
        )
    else:
        return False


def check_role(user: str, required_role: Role, index: str, required_global_role: Role = Role.ADMIN):
    """Check if the given user have at least the given role (in the index, if given), raise Exception otherwise.

    :param user: The email address of the authenticated user
    :param required_role: The minimum role of the user on the given index
    :param index: The index to check the role on
    :param required_global_role: If the user has this global role (default: admin), also allow them access
    :return: the actual role of the user on this index
    """
    # First, check global role (also checks that user exists and deals with 'admin' special user)
    if check_global_role(user, required_global_role, raise_error=False):
        return get_role(index, user)
    # Global role check was false, so now check local role
    actual_role = get_role(index, user)

    if get_settings().auth == AuthOptions.no_auth:
        return actual_role
    elif actual_role and actual_role >= required_role:
        return actual_role
    else:
        raise HTTPException(
            status_code=401,
            detail=f"User {user} does not have " f"{required_role.name.title()} permissions on index {index}",
        )


def check_fields_access(index: str, user: str, fields: list[FieldSpec]) -> None:
    """Check if the given user is allowed to query the given fields and snippets on the given index.

    :param index: The index to check the role on
    :param user: The email address of the authenticated user
    :param fields: The fields to check
    :param snippets: The snippets to check
    :return: Nothing. Throws HTTPException if the user is not allowed to query the given fields and snippets.
    """

    role = get_role(index, user)
    if role is None:
        raise HTTPException(
            status_code=401,
            detail=f"User {user} does not have a role on index {index}",
        )
    if role >= Role.READER:
        return None
    if fields is None:
        return None

    # after this, we know the user is a metareader, so we need to check metareader_access
    index_fields = get_fields(index)
    for field in fields:
        if field.name not in index_fields:
            # might be better to raise an error here, but since we want to support querying multiple
            # indices at once, this allows the user to query fields that do not exist on all indices
            continue
        metareader = index_fields[field.name].metareader

        if metareader.access == "read":
            continue
        elif metareader.access == "snippet" and metareader.max_snippet is not None:
            if metareader.max_snippet is None:
                max_params_msg = ""
            else:
                max_params_msg = (
                    "Can only read snippet with max parameters:"
                    f" nomatch_chars={metareader.max_snippet.nomatch_chars}"
                    f", max_matches={metareader.max_snippet.max_matches}"
                    f", match_chars={metareader.max_snippet.match_chars}"
                )
            if field.snippet is None:
                # if snippet is not specified, the whole field is requested
                raise HTTPException(
                    status_code=401, detail=f"METAREADER cannot read {field} on index {index}. {max_params_msg}"
                )

            valid_nomatch_chars = field.snippet.nomatch_chars <= metareader.max_snippet.nomatch_chars
            valid_max_matches = field.snippet.max_matches <= metareader.max_snippet.max_matches
            valid_match_chars = field.snippet.match_chars <= metareader.max_snippet.match_chars
            valid = valid_nomatch_chars and valid_max_matches and valid_match_chars
            if not valid:
                raise HTTPException(
                    status_code=401,
                    detail=f"The requested snippet of {field.name} on index {index} is too long. {max_params_msg}",
                )
        else:
            raise HTTPException(
                status_code=401,
                detail=f"METAREADER cannot read {field.name} on index {index}",
            )


async def authenticated_user(token: str = Depends(oauth2_scheme)) -> str:
    """Dependency to verify and return a user based on a token."""
    auth = get_settings().auth
    if token is None:
        if auth == AuthOptions.no_auth:
            return ADMIN_USER
        elif auth == AuthOptions.allow_guests:
            return GUEST_USER
        else:
            raise HTTPException(
                status_code=HTTP_401_UNAUTHORIZED,
                detail="This instance has no guest access, please provide a valid bearer token",
            )
    try:
        user = verify_token(token)["email"]
    except Exception:
        logging.exception("Login failed")
        raise HTTPException(status_code=401, detail="Invalid token")
    if auth == AuthOptions.authorized_users_only:
        if get_global_role(user) is None:
            raise HTTPException(
                status_code=401,
                detail=f"The user {user} is not authorized to access this AmCAT instance",
            )
    return user


async def authenticated_writer(user: str = Depends(authenticated_user)):
    """Dependency to verify and return a global writer user based on a token."""
    if get_settings().auth != AuthOptions.no_auth:
        check_global_role(user, Role.WRITER)
    return user


async def authenticated_admin(user: str = Depends(authenticated_user)):
    """Dependency to verify and return a global writer user based on a token."""
    if get_settings().auth != AuthOptions.no_auth:
        check_global_role(user, Role.ADMIN)
    return user



import itertools
from typing import Optional
from fastapi import APIRouter, Depends, HTTPException

from amcat4 import index, multimedia
from amcat4.api.auth import authenticated_user, check_role
from minio.datatypes import Object
from minio.error import S3Error

app_multimedia = APIRouter(prefix="/index/{ix}/multimedia", tags=["multimedia"])


@app_multimedia.get("/presigned_get")
def presigned_get(ix: str, key: str, user: str = Depends(authenticated_user)):
    check_role(user, index.Role.READER, ix)
    try:
        url = multimedia.presigned_get(ix, key)
        obj = multimedia.stat_multimedia_object(ix, key)
        return dict(url=url, content_type=(obj.content_type,), size=obj.size)
    except S3Error as e:
        if e.code == "NoSuchKey":
            raise HTTPException(status_code=404, detail=f"multimedia file {key} not found")
        raise HTTPException(status_code=404, detail=e.message)


@app_multimedia.get("/presigned_post")
def presigned_post(ix: str, user: str = Depends(authenticated_user)):
    check_role(user, index.Role.WRITER, ix)
    url, form_data = multimedia.presigned_post(ix)
    return dict(url=url, form_data=form_data)


@app_multimedia.get("/list")
def list_multimedia(
    ix: str,
    n: int = 10,
    prefix: Optional[str] = None,
    start_after: Optional[str] = None,
    recursive=False,
    presigned_get=False,
    metadata=False,
    user: str = Depends(authenticated_user),
):
    recursive = str(recursive).lower() == "true"
    metadata = str(metadata).lower() == "true"
    presigned_get = str(presigned_get).lower() == "true"

    def process(obj: Object):
        if metadata and (not obj.is_dir) and obj.object_name:
            obj = multimedia.stat_multimedia_object(ix, obj.object_name)
        result: dict[str, object] = dict(
            key=obj.object_name,
            is_dir=obj.is_dir,
            last_modified=obj.last_modified,
            size=obj.size,
        )
        if metadata:
            result["metadata"] = (obj.metadata,)
            result["content_type"] = (obj.content_type,)

        if presigned_get is True and not obj.is_dir:
            if n > 10:
                raise ValueError("Cannot provide presigned_get for more than 10 objects")
            result["presigned_get"] = multimedia.presigned_get(ix, obj.object_name)
        return result

    check_role(user, index.Role.READER, ix)
    objects = multimedia.list_multimedia_objects(ix, prefix, start_after, recursive)
    return [process(obj) for obj in itertools.islice(objects, n)]



"""
All things query
"""

from math import ceil

from typing import (
    Union,
    Sequence,
    Any,
    Dict,
    Tuple,
    Literal,
)

from urllib3 import Retry

from amcat4.models import FieldSpec, FilterSpec, SortSpec

from .date_mappings import mappings
from .elastic import es
from amcat4.index import update_documents_by_query, update_tag_by_query


def build_body(
    queries: dict[str, str] | None = None,
    filters: dict[str, FilterSpec] | None = None,
    highlight: dict | None = None,
    ids: list[str] | None = None,
):
    def parse_filter(field: str, filterSpec: FilterSpec) -> Tuple[dict, dict]:
        filter = filterSpec.model_dump(exclude_none=True)
        extra_runtime_mappings = {}
        field_filters = []
        for value in filter.pop("values", []):
            field_filters.append({"term": {field: value}})
        if "value" in filter:
            field_filters.append({"term": {field: filter.pop("value")}})
        if "exists" in filter:
            if filter.pop("exists"):
                field_filters.append({"exists": {"field": field}})
            else:
                field_filters.append({"bool": {"must_not": {"exists": {"field": field}}}})
        for mapping in mappings():
            if mapping.interval in filter:
                value = filter.pop(mapping.interval)
                extra_runtime_mappings.update(mapping.mapping(field))
                field_filters.append({"term": {mapping.fieldname(field): value}})
        rangefilter = {}
        for rangevar in ["gt", "gte", "lt", "lte"]:
            if rangevar in filter:
                rangefilter[rangevar] = filter.pop(rangevar)
        if rangefilter:
            field_filters.append({"range": {field: rangefilter}})
        if filter:
            raise ValueError(f"Unknown filter type(s): {filter}")
        return extra_runtime_mappings, {"bool": {"should": field_filters}}

    def parse_query(q: str) -> dict:
        return {"query_string": {"query": q}}

    def parse_queries(queries: dict[str, str]) -> dict:
        qs = queries.values()
        if len(qs) == 1:
            return parse_query(list(qs)[0])
        else:
            return {"bool": {"should": [parse_query(q) for q in qs]}}

    if not (queries or filters or ids or highlight):
        return {"query": {"match_all": {}}}

    fs, runtime_mappings = [], {}
    if filters:
        for field, filter in filters.items():
            extra_runtime_mappings, filter_term = parse_filter(field, filter)
            fs.append(filter_term)
            if extra_runtime_mappings:
                runtime_mappings.update(extra_runtime_mappings)
    if queries is not None:
        fs.append(parse_queries(queries))
    if ids:
        fs.append({"ids": {"values": list(ids)}})
    body: Dict[str, Any] = {"query": {"bool": {"filter": fs}}}
    if runtime_mappings:
        body["runtime_mappings"] = runtime_mappings

    if highlight is not None:
        body["highlight"] = highlight

    return body


class QueryResult:
    def __init__(
        self,
        data: list[dict],
        n: int | None = None,
        per_page: int | None = None,
        page: int | None = None,
        page_count: int | None = None,
        scroll_id: str | None = None,
    ):
        if n and (page_count is None) and (per_page is not None):
            page_count = ceil(n / per_page)
        self.data = data
        self.total_count = n
        self.page = page
        self.page_count = page_count
        self.per_page = per_page
        self.scroll_id = scroll_id

    def as_dict(self) -> dict:
        meta: dict[str, int | str | None] = {
            "total_count": self.total_count,
            "per_page": self.per_page,
            "page_count": self.page_count,
        }
        if self.scroll_id:
            meta["scroll_id"] = self.scroll_id
        else:
            meta["page"] = self.page
        return dict(meta=meta, results=self.data)


def query_documents(
    index: Union[str, list[str]],
    fields: list[FieldSpec] | None = None,
    queries: dict[str, str] | None = None,
    filters: dict[str, FilterSpec] | None = None,
    sort: list[dict[str, SortSpec]] | None = None,
    *,
    page: int = 0,
    per_page: int = 10,
    scroll=None,
    scroll_id: str | None = None,
    highlight: bool = False,
    **kwargs,
) -> QueryResult | None:
    """
    Conduct a query_string query, returning the found documents.

    It will return at most per_page results.
    In normal (paginated) mode, the next batch can be  requested by incrementing the page parameter.
    If the scroll parameter is given, the result will contain a scroll_id which can be used to get the next batch.
    In case there are no more documents to scroll, it will return None
    :param index: The name of the index or indexes
    :param fields: List of fields using the FieldSpec syntax. If not specified, only return _id.
                   !Any logic for determining whether a user can see the field should be done in the API layer.
    :param queries: if not None, a dict with labels and queries {label1: query1, ...}
    :param filters: if not None, a dict where the key is the field and the value is a FilterSpec

    :param page: The number of the page to request (starting from zero)
    :param per_page: The number of hits per page
    :param scroll: if not None, will create a scroll request rather than a paginated request. Parmeter should
                   specify the time the context should be kept alive, or True to get the default of 2m.
    :param scroll_id: if not None, should be a previously returned context_id to retrieve a new page of results
    :param highlight: if True, add <em> tags to query matches in fields
    :param sort: Sort order of results, can be either a single field or a list of fields.
                 In the list, each field is a string or a dict with options, e.g. ["id", {"date": {"order": "desc"}}]
                 (https://www.elastic.co/guide/en/elasticsearch/reference/current/sort-search-results.html)
    :param kwargs: Additional elements passed to Elasticsearch.search()
    :return: a QueryResult, or None if there is not scroll result anymore
    """
    if fields is not None and not isinstance(fields, list):
        raise ValueError("fields should be a list")

    if scroll or scroll_id:
        # set scroll to default also if scroll_id is given but no scroll time is known
        kwargs["scroll"] = "2m" if (not scroll or scroll is True) else scroll

    if sort is not None:
        kwargs["sort"] = []
        for s in sort:
            for k, v in s.items():
                kwargs["sort"].append({k: dict(v)})

    if scroll_id:
        result = es().scroll(scroll_id=scroll_id, **kwargs)
        if not result["hits"]["hits"]:
            return None
    else:
        h = query_highlight(fields, highlight) if fields is not None else None
        body = build_body(queries, filters, h)

        fieldnames = [field.name for field in fields] if fields is not None else ["_id"]
        kwargs["_source"] = fieldnames

        if not scroll:
            kwargs["from_"] = page * per_page
        result = es().search(index=index, size=per_page, **body, **kwargs)

    data = []
    for hit in result["hits"]["hits"]:
        hitdict = dict(_id=hit["_id"], **hit["_source"])
        hitdict = overwrite_highlight_results(hit, hitdict)
        if "highlight" in hit:
            for key in hit["highlight"].keys():
                if hit["highlight"][key]:
                    hitdict[key] = " ... ".join(hit["highlight"][key])
        data.append(hitdict)
    if scroll_id:
        return QueryResult(data, n=result["hits"]["total"]["value"], scroll_id=result["_scroll_id"])
    elif scroll:
        return QueryResult(
            data,
            n=result["hits"]["total"]["value"],
            per_page=per_page,
            scroll_id=result["_scroll_id"],
        )
    else:
        n = result["hits"]["total"]["value"]
        if n == 10000:
            # Default elastic max on non-scrolled values. I think we should return the actual count,
            # even if elastic will error (I think) if the user ever retrieves a page > 1000
            # TODO: can we not hard-code the 10k limit?
            res = es().count(index=index, query=body["query"])
            n = res["count"]

        return QueryResult(data, n=n, per_page=per_page, page=page)


def query_highlight(fields: list[FieldSpec], highlight_queries: bool = False) -> dict[str, Any]:
    """
    The elastic "highlight" parameters works for both highlighting text fields and adding snippets.
    This function will return the highlight parameter to be added to the query body.
    """

    highlight: dict[str, Any] = {
        "pre_tags": ["<em>"] if highlight_queries is True else [""],
        "post_tags": ["</em>"] if highlight_queries is True else [""],
        "require_field_match": True,
        "fields": {},
    }

    for field in fields:
        if field.snippet is None:
            if highlight_queries is True:
                # This will overwrite the field with the highlighted version, so
                # only needed if highlight is True
                highlight["fields"][field.name] = {"number_of_fragments": 0}
        else:
            # the elastic highlight feature is also used to get snippets. note that
            # above in the
            highlight["fields"][field.name] = {
                "no_match_size": field.snippet.nomatch_chars,
                "number_of_fragments": field.snippet.max_matches,
                "fragment_size": field.snippet.match_chars or 1,  # 0 would return the whole field
            }
            if field.snippet.max_matches == 0:
                # If max_matches is zero, we drop the query for highlighting so that
                # the nomatch_chars are returned
                highlight["fields"][field.name]["highlight_query"] = {"match_all": {}}

    return highlight


def overwrite_highlight_results(hit: dict, hitdict: dict):
    """
    highlights are a separate field in the hits. If highlight is True, we want to overwrite
    the original field with the highlighted version. If there are snippets, we want to add them
    """
    if not hit.get("highlight"):
        return hitdict
    for key in hit["highlight"].keys():
        if hit["highlight"][key]:
            hitdict[key] = " ... ".join(hit["highlight"][key])
    return hitdict


def update_tag_query(
    index: str | list[str],
    action: Literal["add", "remove"],
    field: str,
    tag: str,
    queries: dict[str, str] | None = None,
    filters: dict[str, FilterSpec] | None = None,
    ids: list[str] | None = None,
):
    """Add or remove tags using a query"""
    body = build_body(queries, filters, ids=ids)

    update_result = update_tag_by_query(index, action, body, field, tag)
    return update_result


def update_query(
    index: str | list[str],
    field: str,
    value: Any,
    queries: dict[str, str] | None = None,
    filters: dict[str, FilterSpec] | None = None,
    ids: list[str] | None = None,
):
    query = build_body(queries, filters, ids=ids)
    return update_documents_by_query(index=index, query=query["query"], field=field, value=value)



from curses import OK
from xml.dom.domreg import registered
import pydantic
from pydantic import BaseModel, field_validator, model_validator, validator
from typing import Annotated, Any, Literal
from typing_extensions import Self


FieldType = Literal[
    "text",
    "date",
    "boolean",
    "keyword",
    "number",
    "integer",
    "object",
    "vector",
    "geo_point",
    "image",
    "video",
    "audio",
    "tag",
    "json",
    "url",
    "preprocess",
]
ElasticType = Literal[
    "text",
    "annotated_text",
    "binary",
    "match_only_text",
    "date",
    "boolean",
    "keyword",
    "constant_keyword",
    "wildcard",
    "integer",
    "byte",
    "short",
    "long",
    "unsigned_long",
    "float",
    "half_float",
    "double",
    "scaled_float",
    "object",
    "flattened",
    "nested",
    "dense_vector",
    "geo_point",
]


class SnippetParams(BaseModel):
    """
    Snippet parameters for a specific field.
    nomatch_chars is the number of characters to show if there is no query match. This is always
    the first [nomatch_chars] of the field.
    """

    nomatch_chars: Annotated[int, pydantic.Field(ge=1)] = 100
    max_matches: Annotated[int, pydantic.Field(ge=0)] = 0
    match_chars: Annotated[int, pydantic.Field(ge=1)] = 50


class FieldMetareaderAccess(BaseModel):
    """Metareader access for a specific field."""

    access: Literal["none", "read", "snippet"] = "none"
    max_snippet: SnippetParams | None = None


class Field(BaseModel):
    """Settings for a field. Some settings, such as metareader, have a strict type because they are used
    server side. Others, such as client_settings, are free-form and can be used by the client to store settings."""

    type: FieldType
    elastic_type: ElasticType
    identifier: bool = False
    metareader: FieldMetareaderAccess = FieldMetareaderAccess()
    client_settings: dict[str, Any] = {}

    @model_validator(mode="after")
    def validate_type(self) -> Self:
        if self.identifier:
            # Identifiers have to be immutable. Instead of checking this in every endpoint that performs updates,
            # we can disable it for certain types that are known to be mutable.
            for forbidden_type in ["tag"]:
                if self.type == forbidden_type:
                    raise ValueError(f"Field type {forbidden_type} cannot be used as an identifier")
        return self


class CreateField(BaseModel):
    """Model for creating a field"""

    type: FieldType
    elastic_type: ElasticType | None = None
    identifier: bool = False
    metareader: FieldMetareaderAccess | None = None
    client_settings: dict[str, Any] | None = None


class UpdateField(BaseModel):
    """Model for updating a field"""

    type: FieldType | None = None
    metareader: FieldMetareaderAccess | None = None
    client_settings: dict[str, Any] | None = None


FilterValue = str | int


class FilterSpec(BaseModel):
    """Form for filter specification."""

    values: list[FilterValue] | None = None
    gt: FilterValue | None = None
    lt: FilterValue | None = None
    gte: FilterValue | None = None
    lte: FilterValue | None = None
    exists: bool | None = None


class FieldSpec(BaseModel):
    """Form for field specification."""

    name: str
    snippet: SnippetParams | None = None


class SortSpec(BaseModel):
    """Form for sort specification."""

    order: Literal["asc", "desc"] = "asc"



"""
User/Account and authentication endpoints.

AmCAT4 can use either Basic or Token-based authentication.
A client can request a token with basic authentication and store that token for future requests.
"""

from typing import Literal, Optional
from importlib.metadata import version

from fastapi import APIRouter, HTTPException, status, Response, Depends
from pydantic import BaseModel
from pydantic.networks import EmailStr

from amcat4 import index
from amcat4.api.auth import authenticated_user, authenticated_admin, check_global_role
from amcat4.config import get_settings, validate_settings
from amcat4.index import ADMIN_USER, GUEST_USER, Role, set_global_role, get_global_role, user_exists

app_users = APIRouter(tags=["users"])


ROLE = Literal["ADMIN", "WRITER", "READER", "NONE"]


class UserForm(BaseModel):
    """Form to create a new global user."""

    email: EmailStr
    role: Optional[ROLE] = None


class ChangeUserForm(BaseModel):
    """Form to change a global user."""

    role: Optional[ROLE] = None


@app_users.post("/users", status_code=status.HTTP_201_CREATED)
def create_user(new_user: UserForm, _=Depends(authenticated_admin)):
    """Create a new user."""
    if user_exists(new_user.email):
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=f"User {new_user.email} already exists",
        )
    if new_user.role is None:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=f"User requires a role (one of {', '.join([r.name for r in Role])})",
        )
    role = Role[new_user.role.upper()]
    set_global_role(email=new_user.email, role=role)
    return {"email": new_user.email, "global_role": role.value}


@app_users.get("/users/me")
def get_current_user(current_user: str = Depends(authenticated_user)):
    """View the current user."""
    return _get_user(current_user, current_user)


@app_users.get("/users/{email}")
def get_user(email: EmailStr, current_user: str = Depends(authenticated_user)):
    """
    View a specified current user.

    Users can view themselves, writer can view others
    """
    return _get_user(email, current_user)


def _get_user(email, current_user):
    if current_user != email:
        check_global_role(current_user, Role.WRITER)
    global_role = get_global_role(email)
    if email in (ADMIN_USER, GUEST_USER) or global_role is Role.NONE:
        raise HTTPException(404, detail=f"User {email} unknown")
    else:
        return {"email": email, "role": global_role.name}


@app_users.get("/users", dependencies=[Depends(authenticated_admin)])
def list_global_users():
    """List all global users"""
    return [{"email": email, "role": role.name} for (email, role) in index.list_global_users().items()]


@app_users.delete("/users/{email}", status_code=status.HTTP_204_NO_CONTENT, response_class=Response)
def delete_user(email: EmailStr, current_user: str = Depends(authenticated_user)):
    """
    Delete the given user.

    Users can delete themselves and admin can delete everyone
    """
    if current_user != email:
        check_global_role(current_user, Role.ADMIN)
    index.delete_user(email)


@app_users.put("/users/{email}")
def modify_user(email: EmailStr, data: ChangeUserForm, _user: str = Depends(authenticated_admin)):
    """
    Modify the given user.
    Only admin can change users.
    """
    if data.role is None or data.role == "NONE":
        set_global_role(email, None)
        return {"email": email, "role": None}
    else:
        role = Role[data.role.upper()]
        set_global_role(email, role)
        return {"email": email, "role": role.name}


@app_users.get("/config")
@app_users.get("/middlecat")
def get_auth_config():
    return {
        "middlecat_url": get_settings().middlecat_url,
        "resource": get_settings().host,
        "authorization": get_settings().auth,
        "warnings": [validate_settings()],
        "api_version": version("amcat4"),
    }



"""Helper methods for the API."""

from pydantic.main import BaseModel


def py2dict(m: BaseModel) -> dict:
    """Convert a pydantic object to a regular dict."""
    return {k: v for (k, v) in m.model_dump().items() if v is not None}



import functools
from multiprocessing import Value
from re import I
from typing import Any, Dict, List, Literal, Optional
from pydantic import BaseModel
import jsonpath_ng

from amcat4.models import FieldType

"""
https://huggingface.co/docs/api-inference/detailed_parameters

"""


class PreprocessingRequest(BaseModel):
    body: Literal["json", "binary"]
    template: Optional[dict] = None


class PreprocessingSetting(BaseModel):
    name: str
    type: str = "string"
    path: Optional[str] = None

    @functools.cached_property
    def parsed(self) -> jsonpath_ng.JSONPath:
        return jsonpath_ng.parse(self.path)


class PreprocessingOutput(PreprocessingSetting):
    recommended_type: FieldType


class PreprocessingParameter(PreprocessingSetting):
    use_field: Literal["yes", "no"] = "no"
    default: Optional[bool | str | int | float] = None
    placeholder: Optional[str] = None
    header: Optional[bool] = None
    secret: Optional[bool] = False


class PreprocessingEndpoint(BaseModel):
    placeholder: str
    domain: List[str]


class PreprocessingTask(BaseModel):
    """Form for query metadata."""

    name: str
    endpoint: PreprocessingEndpoint
    parameters: List[PreprocessingParameter]
    outputs: List[PreprocessingOutput]
    request: PreprocessingRequest

    def get_parameter(self, name) -> PreprocessingParameter:
        # TODO should probably cache this
        for param in self.parameters:
            if param.name == name:
                return param
        raise ValueError(f"Parameter {name} not defined on task {self.name}")

    def get_output(self, name) -> PreprocessingOutput:
        # TODO should probably cache this
        for output in self.outputs:
            if output.name == name:
                return output
        raise ValueError(f"Parameter {name} not defined on task {self.name}")


TASKS: List[PreprocessingTask] = [
    PreprocessingTask(
        # https://huggingface.co/docs/api-inference/detailed_parameters#zero-shot-classification-task
        name="HuggingFace Zero-Shot",
        endpoint=PreprocessingEndpoint(
            placeholder="https://api-inference.huggingface.co/models/facebook/bart-large-mnli",
            domain=["huggingface.co", "huggingfacecloud.com"],
        ),
        parameters=[
            PreprocessingParameter(name="input", type="string", use_field="yes", path="$.inputs"),
            PreprocessingParameter(
                name="candidate_labels",
                type="string[]",
                use_field="no",
                placeholder="politics, sports",
                path="$.parameters.candidate_labels",
            ),
            PreprocessingParameter(
                name="Huggingface Token",
                type="string",
                use_field="no",
                header=True,
                path="Authorization:Bearer",
                secret=True,
            ),
        ],
        outputs=[PreprocessingOutput(name="label", recommended_type="keyword", path="$.labels[0]")],
        request=PreprocessingRequest(body="json", template={"inputs": "", "parameters": {"candidate_labels": ""}}),
    ),
    PreprocessingTask(
        # https://huggingface.co/docs/api-inference/detailed_parameters#zero-shot-classification-task
        name="HuggingFace Image Classification",
        endpoint=PreprocessingEndpoint(
            placeholder="https://api-inference.huggingface.co/models/google/vit-base-patch16-224",
            domain=["huggingface.co", "huggingfacecloud.com"],
        ),
        parameters=[
            PreprocessingParameter(name="input", type="image", use_field="yes"),
            PreprocessingParameter(
                name="Huggingface Token",
                type="string",
                use_field="no",
                header=True,
                path="Authorization:Bearer",
                secret=True,
            ),
        ],
        outputs=[PreprocessingOutput(name="label", recommended_type="keyword", path="$[0].label")],
        request=PreprocessingRequest(body="binary"),
    ),
]


@functools.cache
def get_task(name):
    for task in TASKS:
        if task.name == name:
            return task
    raise ValueError(f"Task {task} not defined")


def get_tasks():
    return TASKS



from fastapi import Request
from fastapi import APIRouter
from fastapi.templating import Jinja2Templates
from importlib.metadata import version

from amcat4 import elastic
from amcat4.api.auth import get_middlecat_config
from amcat4.config import get_settings

templates = Jinja2Templates(directory="templates")

app_info = APIRouter(tags=["informational"])


@app_info.get("/")
def index(request: Request):
    host = get_settings().host
    es_alive = elastic.ping()
    auth = get_settings().auth
    has_admin_email = bool(get_settings().admin_email)
    middlecat_url = get_settings().middlecat_url
    middlecat_alive = False
    api_version = version("amcat4")
    if middlecat_url:
        try:
            get_middlecat_config(middlecat_url)
            middlecat_alive = True
        except OSError:
            pass
    return templates.TemplateResponse("index.html", locals())



"""AmCAT4 API."""

from contextlib import asynccontextmanager
import logging
from fastapi import FastAPI, Request
from fastapi.responses import JSONResponse
from fastapi.middleware.cors import CORSMiddleware
from fastapi.middleware.gzip import GZipMiddleware

from amcat4.api.index import app_index
from amcat4.api.info import app_info
from amcat4.api.query import app_query
from amcat4.api.users import app_users
from amcat4.api.multimedia import app_multimedia
from amcat4.api.preprocessing import app_preprocessing
from amcat4.preprocessing.processor import start_processors


@asynccontextmanager
async def lifespan(app: FastAPI):
    try:
        start_processors()
    except:
        logging.exception("Error on initializing preprocessing")
    yield


app = FastAPI(
    title="AmCAT4",
    description=__doc__,
    openapi_tags=[
        dict(name="users", description="Endpoints for user management"),
        dict(name="index", description="Endpoints to create, list, and delete indices; and to add or modify documents"),
        dict(name="query", description="Endpoints to list or query documents or run aggregate queries"),
        dict(name="middlecat", description="MiddleCat authentication"),
        dict(name="annotator users", description="Annotator module endpoints for user management"),
        dict(
            name="annotator codingjob",
            description="Annotator module endpoints for creating and managing annotator codingjobs, "
            "and the core process of getting units and posting annotations",
        ),
        dict(name="annotator guest", description="Annotator module endpoints for unregistered guests"),
        dict(name="multimedia", description="Endpoints for multimedia support"),
        dict(name="preprocessing", description="Endpoints for preprocessing support"),
    ],
    lifespan=lifespan,
)
app.include_router(app_info)
app.include_router(app_users)
app.include_router(app_index)
app.include_router(app_query)
app.include_router(app_multimedia)
app.include_router(app_preprocessing)
app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)
app.add_middleware(GZipMiddleware, minimum_size=1000)


@app.exception_handler(ValueError)
async def value_error_exception_handler(request: Request, exc: ValueError):
    return JSONResponse(
        status_code=400,
        content={"message": str(exc)},
    )



"""
AmCAT4 Configuration

We read configuration from 2 sources, in order of precedence (higher is more priority)
- Environment variables
- A .env file, either in the current working directory or in a location specified
  by the AMCAT4_CONFIG_FILE environment variable
"""

import functools
from enum import Enum
from pathlib import Path
from typing import Annotated, Any
from class_doc import extract_docs_from_cls_obj
from dotenv import load_dotenv
from pydantic import model_validator, Field
from pydantic_settings import BaseSettings, SettingsConfigDict

ENV_PREFIX = "amcat4_"


class AuthOptions(str, Enum):
    #: everyone (that can reach the server) can do anything they want
    no_auth = "no_auth"

    #: everyone can use the server, dependent on index-level guest_role authorization settings
    allow_guests = "allow_guests"

    #: everyone can use the server, if they have a valid middlecat login,
    #: and dependent on index-level guest_role authorization settings
    allow_authenticated_guests = "allow_authenticated_guests"

    #: only people with a valid middlecat login and an explicit server role can use the server
    authorized_users_only = "authorized_users_only"

    @classmethod
    def validate(cls, value: str):
        if value not in cls.__members__:
            options = ", ".join(AuthOptions.__members__.keys())
            return f"{value} is not a valid authorization option. Choose one of {{{options}}}"


# Set the __doc__ attribute of each AuthOptions enum member using extract_docs_from_cls_obj
for field, doc in extract_docs_from_cls_obj(AuthOptions).items():
    AuthOptions[field].__doc__ = "\n".join(doc)


class Settings(BaseSettings):
    env_file: Annotated[
        Path,
        Field(
            description="Location of a .env file (if used) relative to working directory",
        ),
    ] = Path(".env")
    host: Annotated[
        str,
        Field(
            description="Host this instance is served at (needed for checking tokens)",
        ),
    ] = "http://localhost:5000"

    elastic_password: Annotated[
        str | None,
        Field(
            description=(
                "Elasticsearch password. " "This the password for the 'elastic' user when Elastic xpack security is enabled"
            )
        ),
    ] = None

    elastic_host: Annotated[
        str | None,
        Field(
            description=(
                "Elasticsearch host. "
                "Default: https://localhost:9200 if elastic_password is set, http://localhost:9200 otherwise"
            )
        ),
    ] = None

    elastic_verify_ssl: Annotated[
        bool | None,
        Field(
            description=(
                "Elasticsearch verify SSL (only used if elastic_password is set). " "Default: True unless host is localhost)"
            ),
        ),
    ] = None

    system_index: Annotated[
        str,
        Field(
            description="Elasticsearch index to store authorization information in",
        ),
    ] = "amcat4_system"

    auth: Annotated[AuthOptions, Field(description="Do we require authorization?")] = AuthOptions.no_auth

    middlecat_url: Annotated[
        str,
        Field(
            description="Middlecat server to trust as ID provider",
        ),
    ] = "https://middlecat.net"

    admin_email: Annotated[
        str | None,
        Field(
            description="Email address for a hardcoded admin email (useful for setup and recovery)",
        ),
    ] = None

    minio_host: Annotated[str | None, Field()] = None
    minio_tls: Annotated[bool, Field()] = False
    minio_access_key: Annotated[str | None, Field()] = None
    minio_secret_key: Annotated[str | None, Field()] = None

    @model_validator(mode="after")
    def set_ssl(self: Any) -> "Settings":
        if not self.elastic_host:
            self.elastic_host = ("https" if self.elastic_password else "http") + "://localhost:9200"
        if not self.elastic_verify_ssl:
            self.elastic_verify_ssl = self.elastic_host not in {
                "http://localhost:9200",
                "https://localhost:9200",
            }
        return self

    model_config = SettingsConfigDict(env_prefix=ENV_PREFIX)


@functools.lru_cache()
def get_settings() -> Settings:
    # This shouldn't be necessary according to the docs, but without the load_dotenv it doesn't work at
    # least when running with python -m amcat4.config...
    temp = Settings()
    # WvA: For some reason, it always seems to override environment variables?
    load_dotenv(temp.env_file, override=False)
    return Settings()


def validate_settings():
    if get_settings().auth != "no_auth":
        if get_settings().host.startswith("http://") and not get_settings().host.startswith("http://localhost"):
            return (
                "You have set the host at an http address and enabled authentication."
                "Authentication through middlecat will not work in your browser"
                " without additional steps. See https://github.com/ccs-amsterdam/amcat4py/issues/9"
            )


if __name__ == "__main__":
    # Echo the settings
    for k, v in get_settings().model_dump().items():
        print(f"{ENV_PREFIX.upper()}{k.upper()}={v}")



"""
Connection between AmCAT4 and the elasticsearch backend

Some things to note:
- See config.py for global settings, including elastic host and system index name
- The elasticsearch backend should contain a system index, which will be created if needed
- The system index contains a 'document' for each used index containing:
  {auth: [{email: role}], guest_role: role}

"""

import functools

import logging
from typing import Optional

from elasticsearch import Elasticsearch, NotFoundError
from amcat4.config import get_settings

SYSTEM_INDEX_VERSION = 1

SYSTEM_MAPPING = {
    "name": {"type": "text"},
    "description": {"type": "text"},
    "roles": {"type": "nested"},
    "summary_field": {"type": "keyword"},
    "guest_role": {"type": "keyword"},
}


class CannotConnectElastic(Exception):
    pass


@functools.lru_cache()
def es() -> Elasticsearch:
    try:
        return _setup_elastic()
    except ValueError as e:
        raise ValueError(f"Cannot connect to elastic {get_settings().elastic_host!r}: {e}")


def connect_elastic() -> Elasticsearch:
    """
    Connect to the elastic server using the system settings
    """
    settings = get_settings()
    if settings.elastic_password:
        host = settings.elastic_host
        if settings.elastic_verify_ssl is None:
            verify_certs = "localhost" in (host or "")
        else:
            verify_certs = settings.elastic_verify_ssl

        return Elasticsearch(
            host,
            basic_auth=("elastic", settings.elastic_password),
            verify_certs=verify_certs,
        )
    else:
        return Elasticsearch(settings.elastic_host or None)


def get_system_version(elastic=None) -> Optional[int]:
    """
    Get the elastic system index version
    """
    # WvA: I don't like this 'circular' import. Should probably reorganize the elastic and index modules
    from amcat4.index import GLOBAL_ROLES

    settings = get_settings()
    if elastic is None:
        elastic = es()
    try:
        r = elastic.get(index=settings.system_index, id=GLOBAL_ROLES, source_includes="version")
    except NotFoundError:
        return None
    return r["_source"]["version"]


def _setup_elastic():
    """
    Check whether we can connect with elastic
    """
    # WvA: I don't like this 'circular' import. Should probably reorganize the elastic and index modules
    from amcat4.index import GLOBAL_ROLES

    settings = get_settings()
    logging.debug(
        f"Connecting with elasticsearch at {settings.elastic_host}, "
        f"password? {'yes' if settings.elastic_password else 'no'} "
    )
    elastic = connect_elastic()
    if not elastic.ping():
        raise CannotConnectElastic(f"Cannot connect to elasticsearch server {settings.elastic_host}")
    if elastic.indices.exists(index=settings.system_index):
        # Check index format version
        if get_system_version(elastic) is None:
            raise CannotConnectElastic(
                f"System index {settings.elastic_host}::{settings.system_index} is corrupted or uses an "
                f"old format. Please repair or migrate before continuing"
            )

    else:
        logging.info(f"Creating amcat4 system index: {settings.system_index}")
        elastic.indices.create(index=settings.system_index, mappings={"properties": SYSTEM_MAPPING})
        elastic.index(
            index=settings.system_index,
            id=GLOBAL_ROLES,
            document=dict(version=SYSTEM_INDEX_VERSION, roles=[]),
        )
    return elastic


def ping():
    """
    Can we reach this elasticsearch server
    """
    try:
        return es().ping()
    except CannotConnectElastic as e:
        logging.error(e)
        return False



"""
AmCAT4 REST API
"""

import argparse
import csv
import io
import json
import logging
import os
from pathlib import Path
import secrets
import sys
from typing import Any
import urllib.request
from enum import Enum
from collections import defaultdict
import elasticsearch.helpers

import uvicorn
from pydantic.fields import FieldInfo
from uvicorn.config import LOGGING_CONFIG


from amcat4 import index
from amcat4.config import get_settings, AuthOptions, validate_settings
from amcat4.elastic import connect_elastic, get_system_version, ping
from amcat4.index import GLOBAL_ROLES, create_index, set_global_role, Role, list_global_users, upload_documents
from amcat4.models import ElasticType, FieldType

SOTU_INDEX = "state_of_the_union"


def upload_test_data() -> str:
    url = "https://raw.githubusercontent.com/ccs-amsterdam/example-text-data/master/sotu.csv"
    url_open = urllib.request.urlopen(url)
    csv.field_size_limit(sys.maxsize)
    csvfile = csv.DictReader(io.TextIOWrapper(url_open, encoding="utf-8"))

    # creates the index info on the sqlite db
    create_index(SOTU_INDEX)

    docs = [
        dict(
            title="{Year}: {President}".format(**row),
            text=row["Text"],
            date=row["Date"],
            president=row["President"],
            year=row["Year"],
            party=row["Party"],
        )
        for row in csvfile
    ]
    fields: dict[str, FieldType] = {
        "text": "text",
        "title": "text",
        "date": "date",
        "president": "keyword",
        "party": "keyword",
        "year": "integer",
    }
    upload_documents(SOTU_INDEX, docs, fields)
    return SOTU_INDEX


def run(args):
    auth = get_settings().auth
    logging.info(f"Starting server at port {args.port}, debug={not args.nodebug}, auth={auth}")
    if auth == AuthOptions.no_auth:
        logging.warning(
            "Warning: No authentication is set up - " "everyone who can access this service can view and change all data"
        )
    if validate_settings():
        logging.warning(validate_settings())
    logging.info(
        "To change server config, create an .env file and/or set environment parameters,\n"
        f"{' '*26}see README.md, amcat4/config.py or .env.example for more information.\n"
        f"{' '*26}You can also run `python -m amcat4 config` to create the .env settings file interactively\n"
    )
    if ping():
        logging.info(f"Connect to elasticsearch {get_settings().elastic_host}")
    log_config = "logging.yml" if Path("logging.yml").exists() else LOGGING_CONFIG
    uvicorn.run("amcat4.api:app", host="0.0.0.0", reload=not args.nodebug, port=args.port, log_config=log_config)


def val(val_or_list):
    if isinstance(val_or_list, list):
        if len(val_or_list) == 1:
            return val_or_list[0]
        raise ValueError(f"Cannot extract single value from {val_or_list}")
    return val_or_list


def migrate_index(_args) -> None:
    settings = get_settings()
    elastic = connect_elastic()
    if not elastic.ping():
        logging.error(f"Cannot connect to elasticsearch server {settings.elastic_host}")
        sys.exit(1)
    if not elastic.indices.exists(index=settings.system_index):
        logging.info("System index does not exist yet. It will be created automatically if you run the server")
        sys.exit(1)
    # Check index format version
    version = get_system_version(elastic)
    logging.info(f"{settings.elastic_host}::{settings.system_index} is at version {version or 0}")
    if version == 1:
        logging.info("Nothing to do")
    else:
        logging.info("Migrating to version 1")
        fields = ["index", "email", "role"]
        indices: defaultdict[str, dict[str, str]] = defaultdict(dict)

        for entry in elasticsearch.helpers.scan(elastic, index=settings.system_index, fields=fields, _source=False):
            index, email, role = [val(entry["fields"][field]) for field in fields]
            indices[index][email] = role
        if GLOBAL_ROLES not in indices:
            indices[GLOBAL_ROLES] = {}
        try:
            elastic.indices.delete(index=settings.system_index)
            for index, roles_dict in indices.items():
                guest_role = roles_dict.pop("_guest", None)
                roles_dict.pop("ADMIN", None)
                roles = [{"email": email, "role": role} for (email, role) in roles_dict.items()]
                doc: dict[str, Any] = dict(name=index, guest_role=guest_role, roles=roles)
                if index == GLOBAL_ROLES:
                    doc["version"] = 1
                elastic.index(index=settings.system_index, id=index, document=doc)
        except Exception:
            try:
                open("roles.csv", "w").write(json.dumps(indices, indent=2))
                dest = "roles.csv"
            except Exception:
                print(json.dumps(indices, indent=2))
                dest = "screen"
            logging.exception(
                "Something went wrong in migrating, and the old system index is probably lost. Sorry!"
                f"The authorization information is written to {dest}, I hope this can be fixed..."
            )
            sys.exit(1)


def base_env():
    return dict(
        amcat4_secret_key=secrets.token_hex(nbytes=32),
        amcat4_middlecat_url="https://middlecat.net",
    )


def create_env(args):
    if os.path.exists(".env"):
        print("*** File .env already exists, quitting ***")
        sys.exit(1)

    env = base_env()
    if args.admin_email:
        env["amcat4_admin_email"] = args.admin_email
    with open(".env", "w") as f:
        for key, val in env.items():
            f.write(f"{key}={val}\n")
    os.chmod(".env", 0o600)
    print("*** Created .env file ***")


def create_test_index(_args):
    logging.info("**** Creating test index {} ****".format(SOTU_INDEX))
    index.delete_index(SOTU_INDEX, ignore_missing=True)
    upload_test_data()


def add_admin(args):
    logging.info(f"**** Setting {args.email} to ADMIN ****")
    set_global_role(args.email, Role.ADMIN)


def list_users(_args):
    users = list_global_users()

    # sorted changes the output type of list_global_users?
    # users = sorted(list_global_users(), key=lambda ur: (ur[1], ur[0]))
    if users:
        for user, role in users.items():
            print(f"{role.name:10}: {user}")
    if not users:
        print("(No users defined yet, use add-admin to add users by email)")


def config_amcat(args):
    settings = get_settings()
    # Not a useful entry in an actual env_file
    print(f"Reading/writing settings from {settings.env_file}")
    for fieldname, fieldinfo in settings.model_fields.items():
        if fieldname == "env_file":
            continue

        validation_function = AuthOptions.validate if fieldname == "auth" else None
        value = getattr(settings, fieldname)
        value = menu(fieldname, fieldinfo, value, validation_function=validation_function)
        if value is ABORTED:
            return
        if value is not UNCHANGED:
            setattr(settings, fieldname, value)

    with settings.env_file.open("w") as f:
        for fieldname, fieldinfo in settings.model_fields.items():
            value = getattr(settings, fieldname)
            if doc := fieldinfo.description:
                f.write(f"# {doc}\n")
            if _isenum(fieldinfo) and fieldinfo.annotation:
                f.write("# Valid options:\n")
                for option in fieldinfo.annotation:
                    doc = option.__doc__.replace("\n", " ")
                    f.write(f"# - {option.name}: {doc}\n")
            if val is None:
                f.write(f"#amcat4_{fieldname}=\n\n")
            else:
                f.write(f"amcat4_{fieldname}={value}\n\n")
    os.chmod(".env", 0o600)
    print(f"*** Written {bold('.env')} file to {settings.env_file} ***")


def bold(x):
    return "\033[1m" + str(x) + "\033[0m"


ABORTED = object()
UNCHANGED = object()


def _isenum(fieldinfo: FieldInfo) -> bool:
    try:
        return issubclass(fieldinfo.annotation, Enum) if fieldinfo.annotation is not None else False
    except TypeError:
        return False


def menu(fieldname: str, fieldinfo: FieldInfo, value, validation_function=None):
    print(f"\n{bold(fieldname)}: {fieldinfo.description}")
    if _isenum(fieldinfo) and fieldinfo.annotation:
        print("  Possible choices:")
        options: Any = fieldinfo.annotation
        for option in options:
            print(f"  - {option.name}: {option.__doc__}")
        print()
    print(f"The current value for {bold(fieldname)} is {bold(value)}.")
    while True:
        try:
            value = input("Enter a new value, press [enter] to leave unchanged, or press [control+c] to abort: ")
        except KeyboardInterrupt:
            return ABORTED
        if not value.strip():
            return UNCHANGED
        if validation_function and (message := validation_function(value)):
            print(f"\nInvalid value: {message}")
            continue
        return value


def main():
    parser = argparse.ArgumentParser(description=__doc__, prog="python -m amcat4")

    subparsers = parser.add_subparsers(dest="action", title="action", help="Action to perform:", required=True)
    p = subparsers.add_parser("run", help="Run the backend API in development mode")
    p.add_argument(
        "--no-debug",
        action="store_true",
        dest="nodebug",
        help="Disable debug mode (useful for testing downstream clients)",
    )
    p.add_argument("-p", "--port", help="Port", default=5000)
    p.set_defaults(func=run)

    p = subparsers.add_parser("create-env", help="Create the .env file with a random secret key")
    p.add_argument("-a", "--admin_email", help="The email address of the admin user.")
    p.set_defaults(func=create_env)

    p = subparsers.add_parser("config", help="Configure amcat4 settings in an interactive menu.")
    p.set_defaults(func=config_amcat)

    p = subparsers.add_parser("add-admin", help="Add a global admin")
    p.add_argument("email", help="The email address of the admin user.")
    p.set_defaults(func=add_admin)

    p = subparsers.add_parser("list-users", help="List global users")
    p.set_defaults(func=list_users)

    p = subparsers.add_parser("create-test-index", help=f"Create the {SOTU_INDEX} test index")
    p.set_defaults(func=create_test_index)

    p = subparsers.add_parser("migrate", help="Migrate the system index to the current version")
    p.set_defaults(func=migrate_index)

    args = parser.parse_args()

    logging.basicConfig(format="[%(levelname)-7s:%(name)-15s] %(message)s", level=logging.INFO)
    es_logger = logging.getLogger("elasticsearch")
    es_logger.setLevel(logging.WARNING)
    args.func(args)


if __name__ == "__main__":
    main()



"""
Index and authorization management

Encapsulates elasticsearch indices and provides authorisation through user roles

Authorisation rules:
- There are 5 increasing levels of authorisation: None, metareader, reader, writer, admin
- The 'admin' user automatically has admin rights everywhere (or globally?)
- Users can have a global role and a role on any index. Every index can also have a guest role
- Global roles:
  - Readers can see which indices exist, but cannot access them
  - Writers can create new projects and users (with at most their own global role)
  - Admins can delete projects and assign themselves a role on any index role
- Index roles:
  - None means an index cannot be viewed or accessed in any way (the index is invisible to the user)
  - Metareader means the user can read all properties, do queries, etc., but cannot read the 'text' attribute
    (this is mostly intended to provide access to metadata fields of copyrighted material)
  - Reader can read all properties, do queries, etc., but cannot make changes
  - Writers can add/delete documents, add/delete users (up to their own level), and make other changes (but not delete)
  - Admins can do whatever they want, including deleting the index
- If a user does not have an explicit role on an index, the guest role (if any) is used
- An unauthorized user can still get guest roles, so it can see any indices with a guest role

Note that these rules are not enforced in this module, they should be enforced by the API!

Elasticsearch implementation
- There is a system index (configurable name, default: amcat4_system)
- This system index contains a 'document' for each index:
    {name: "...", description:"...", guest_role: "...", roles: [{email, role}...]}
- A special _global document defines the global properties for this instance (name, roles)
- We define the mappings (field types) based on existing elasticsearch mappings,
    but use field metadata to define specific fields.
"""

import collections
from curses import meta
from dataclasses import field
from enum import IntEnum
import functools
import logging
from typing import Any, Iterable, Mapping, Optional, Literal

import hashlib
import json

import elasticsearch.helpers
from elasticsearch import NotFoundError

# from amcat4.api.common import py2dict
from amcat4.config import AuthOptions, get_settings
from amcat4.elastic import es
from amcat4.fields import (
    coerce_type,
    create_fields,
    create_or_verify_tag_field,
    get_fields,
)
from amcat4.models import CreateField, Field, FieldType
from amcat4.preprocessing.models import PreprocessingInstruction
from amcat4.preprocessing import processor


class Role(IntEnum):
    NONE = 0
    METAREADER = 10
    READER = 20
    WRITER = 30
    ADMIN = 40


class GuestRole(IntEnum):
    NONE = 0
    METAREADER = 10
    READER = 20
    WRITER = 30


ADMIN_USER = "_admin"
GUEST_USER = "_guest"
GLOBAL_ROLES = "_global"

Index = collections.namedtuple(
    "Index",
    ["id", "name", "description", "guest_role", "archived", "roles", "summary_field"],
)


class IndexDoesNotExist(ValueError):
    pass


def refresh_index(index: str):
    """
    Refresh the elasticsearch index
    """
    es().indices.refresh(index=index)


def refresh_system_index():
    """
    Refresh the elasticsearch index
    """
    es().indices.refresh(index=get_settings().system_index)


def list_known_indices(email: str | None = None) -> Iterable[Index]:
    """
    List all known indices, e.g. indices registered in this amcat4 instance
    :param email: if given, only list indices visible to this user
    """
    # TODO: Maybe this can be done with a smart query, rather than getting all indices and filtering in python?
    # I tried the following but had no luck
    # q_guest = {"bool" : {"filter": {"exists": {"field": "guest_role"}},
    #                      "must_not": {"term": {"guest_role": {"value": "none", "case_insensitive": True}}}}}
    # q_role = {"nested": {"path": "roles", "query": {"term": {"roles.email": email}}}}
    # query = {"bool": {"should": [q_guest, q_role]}}
    check_role = not (email is None or get_global_role(email) == Role.ADMIN or get_settings().auth == "no_auth")
    for index in elasticsearch.helpers.scan(es(), index=get_settings().system_index, fields=[], _source=True):
        ix = _index_from_elastic(index)
        if ix.name == GLOBAL_ROLES:
            continue
        if (not check_role) or (ix.guest_role) or (email in ix.roles):
            yield ix


def _index_from_elastic(index):
    src = index["_source"]
    guest_role = src.get("guest_role", "NONE")

    return Index(
        id=index["_id"],
        name=src.get("name", index["_id"]),
        description=src.get("description"),
        guest_role=Role[guest_role] if guest_role in Role.__members__ else Role.NONE,
        archived=src.get("archived"),
        roles=_roles_from_elastic(src.get("roles", [])),
        summary_field=src.get("summary_field"),
    )


def get_index(index: str) -> Index:
    try:
        index_resp = es().get(index=get_settings().system_index, id=index)
    except NotFoundError:
        raise IndexDoesNotExist(index)
    return _index_from_elastic(index_resp)


def create_index(
    index: str,
    guest_role: Optional[Role] = None,
    name: Optional[str] = None,
    description: Optional[str] = None,
    admin: Optional[str] = None,
) -> None:
    """
    Create a new index in elasticsearch and register it with this AmCAT instance
    """
    try:
        get_index(index)
        raise ValueError(f'Index "{index}" already exists')
    except IndexDoesNotExist:
        pass

    es().indices.create(index=index, mappings={"properties": {}})
    register_index(
        index,
        guest_role=guest_role or Role.NONE,
        name=name or index,
        description=description or "",
        admin=admin,
    )


def register_index(
    index: str,
    guest_role: Optional[Role] = None,
    name: Optional[str] = None,
    description: Optional[str] = None,
    admin: Optional[str] = None,
) -> None:
    """
    Register an existing elastic index with this AmCAT instance, i.e. create an entry in the system index
    """
    if not es().indices.exists(index=index):
        raise ValueError(f"Index {index} does not exist")
    system_index = get_settings().system_index
    if es().exists(index=system_index, id=index):
        raise ValueError(f"Index {index} is already registered")
    roles = [dict(email=admin, role="ADMIN")] if admin else []

    es().index(
        index=system_index,
        id=index,
        document=dict(
            name=(name or index),
            roles=roles,
            description=description,
            guest_role=guest_role.name if guest_role is not None else "NONE",
        ),
    )
    refresh_index(system_index)


def delete_index(index: str, ignore_missing=False) -> None:
    """
    Delete an index from AmCAT and deregister it from this instance
    :param index: The name of the index
    :param ignore_missing: If True, do not throw exception if index does not exist
    """
    _es = es().options(ignore_status=404) if ignore_missing else es()
    _es.indices.delete(index=index)
    deregister_index(index, ignore_missing=ignore_missing)


def deregister_index(index: str, ignore_missing=False) -> None:
    """
    Deregister an existing elastic index from this AmCAT instance, i.e. remove all authorization entries
    :param index: The name of the index
    :param ignore_missing: If True, do not throw exception if index does not exist
    """
    system_index = get_settings().system_index
    try:
        es().delete(index=system_index, id=index)
    except NotFoundError:
        if not ignore_missing:
            raise
    else:
        refresh_index(system_index)
    # Stop preprocessing loops on this index
    from amcat4.preprocessing.processor import get_manager

    get_manager().remove_index_preprocessors(index)


def _roles_from_elastic(roles: list[dict]) -> dict[str, Role]:
    return {role["email"]: Role[role["role"].upper()] for role in roles}


def _roles_to_elastic(roles: dict) -> list[dict]:
    return [{"email": email, "role": role.name} for (email, role) in roles.items()]


def set_role(index: str, email: str, role: Optional[Role]):
    """
    Set the role for this user on the given index)
    If role is None, remove the role
    """
    # TODO: It would probably be better to do this with a query script on elastic
    system_index = get_settings().system_index
    try:
        d = es().get(index=system_index, id=index, source_includes="roles")
    except NotFoundError:
        raise ValueError(f"Index {index} is not registered")
    roles_dict = _roles_from_elastic(d["_source"].get("roles", []))
    if role:
        roles_dict[email] = role
    else:
        if email not in roles_dict:
            return  # Nothing to change
        del roles_dict[email]

    es().update(
        index=system_index,
        id=index,
        doc=dict(roles=_roles_to_elastic(roles_dict)),
    )


def set_global_role(email: str, role: Role | None):
    """
    Set the global role for this user
    """
    set_role(index=GLOBAL_ROLES, email=email, role=role)


def set_guest_role(index: str, guest_role: Optional[GuestRole]):
    """
    Set the guest role for this index. Set to None to disallow guest access
    """
    modify_index(index, guest_role=GuestRole.NONE if guest_role is None else guest_role)


def modify_index(
    index: str,
    name: Optional[str] = None,
    description: Optional[str] = None,
    guest_role: Optional[GuestRole] = None,
    archived: Optional[str] = None,
):
    doc = dict(
        name=name,
        description=description,
        guest_role=guest_role.name if guest_role is not None else None,
        archived=archived,
    )

    doc = {x: v for (x, v) in doc.items() if v is not None}
    if doc:
        es().update(index=get_settings().system_index, id=index, doc=doc)


def remove_role(index: str, email: str):
    """
    Remove the role of this user on the given index
    """
    set_role(index, email, role=None)


def remove_global_role(email: str):
    """
    Remove the global role of this user
    """
    remove_role(index=GLOBAL_ROLES, email=email)


def user_exists(email: str, index: str = GLOBAL_ROLES) -> bool:
    """
    Check if a user exists on server (GLOBAL_ROLES) or in a specific index
    """
    try:
        doc = es().get(
            index=get_settings().system_index,
            id=index,
            source_includes=["roles", "guest_role"],
        )
    except NotFoundError:
        raise IndexDoesNotExist(f"Index {index} does not exist or is not registered")
    roles_dict = _roles_from_elastic(doc["_source"].get("roles", []))
    return email in roles_dict


def get_role(index: str, email: str) -> Role:
    """
    Retrieve the role of this user on this index, or the guest role if user has no role
    Raises a ValueError if the index does not exist
    :returns: a Role object, or Role.NONE if the user has no role and no guest role exists
    """
    try:
        doc = es().get(
            index=get_settings().system_index,
            id=index,
            source_includes=["roles", "guest_role"],
        )
    except NotFoundError:
        raise IndexDoesNotExist(f"Index {index} does not exist or is not registered")
    roles_dict = _roles_from_elastic(doc["_source"].get("roles", []))
    if role := roles_dict.get(email):
        return role
    if index == GLOBAL_ROLES:
        return Role.NONE

    # are guests allowed?
    if get_settings().auth == AuthOptions.authorized_users_only:
        # only allow guests if authorized at server level
        global_role = get_global_role(email, only_es=True)
        if global_role == Role.NONE:
            return Role.NONE

    return get_guest_role(index)


def get_guest_role(index: str) -> Role:
    """
    Return the guest role for this index, raising a IndexDoesNotExist if the index does not exist
    :returns: a Role object, or None if global role was NONE
    """
    try:
        d = es().get(
            index=get_settings().system_index,
            id=index,
            source_includes="guest_role",
        )
    except NotFoundError:
        raise IndexDoesNotExist(index)
    role = d["_source"].get("guest_role")
    if role and role in Role.__members__:
        return Role[role]
    return Role.NONE


def get_global_role(email: str, only_es: bool = False) -> Role:
    """
    Retrieve the global role of this user

    :returns: a Role object, or None if the user has no role
    """
    # The 'admin' user is given to everyone in the no_auth scenario
    if only_es is False:
        if email == get_settings().admin_email or email == ADMIN_USER:
            return Role.ADMIN
    return get_role(index=GLOBAL_ROLES, email=email)


def list_users(index: str) -> dict[str, Role]:
    """ "
    List all users and their roles on the given index
    :param index: The index to list roles for.
    :returns: an iterable of (user, Role) pairs
    """
    r = es().get(index=get_settings().system_index, id=index, source_includes="roles")
    return _roles_from_elastic(r["_source"].get("roles", []))


def list_global_users() -> dict[str, Role]:
    """ "
    List all global users and their roles
    :returns: an iterable of (user, Role) pairs
    """
    return list_users(index=GLOBAL_ROLES)


def delete_user(email: str) -> None:
    """Delete this user from all indices"""
    set_global_role(email, None)
    for ix in list_known_indices(email):
        set_role(ix.id, email, None)


def create_id(document: dict, field_settings: dict[str, Field]) -> str:
    """
    Create the _id for a document.
    """

    identifiers = [k for k, v in field_settings.items() if v.identifier == True]
    if len(identifiers) == 0:
        raise ValueError("Can only create id if identifiers are specified")

    id_keys = sorted(set(identifiers) & set(document.keys()))
    id_fields = {k: document[k] for k in id_keys}
    hash_str = json.dumps(id_fields, sort_keys=True, ensure_ascii=True, default=str).encode("ascii")
    m = hashlib.sha224()
    m.update(hash_str)
    return m.hexdigest()


def upload_documents(
    index: str,
    documents: list[dict[str, Any]],
    fields: Mapping[str, FieldType | CreateField] | None = None,
    op_type: Literal["create", "update"] = "create",
    raise_on_error=False,
):
    """
    Upload documents to this index

    :param index: The name of the index (without prefix)
    :param documents: A sequence of article dictionaries
    :param fields: A mapping of fieldname:UpdateField for field types
    :param op_type: Whether to 'index' new documents (default) or 'update' existing documents
    """
    if fields:
        create_fields(index, fields)

    def es_actions(index, documents, op_type):
        field_settings = get_fields(index)
        has_identifiers = any(field.identifier for field in field_settings.values())
        for document in documents:
            doc = dict()
            action = {"_op_type": op_type, "_index": index}

            for key in document.keys():
                if key in field_settings:
                    doc[key] = coerce_type(document[key], field_settings[key].type)
                else:
                    if key != "_id":
                        raise ValueError(f"Field '{key}' is not yet specified")

                if key == "_id":
                    if has_identifiers:
                        identifiers = ", ".join([name for name, field in field_settings.items() if field.identifier])
                        raise ValueError(f"This index uses identifier ({identifiers}), so you cannot set the _id directly.")
                    action["_id"] = document["_id"]
                else:
                    if has_identifiers:
                        action["_id"] = create_id(document, field_settings)
                ## if no id is given, elasticsearch creates a cool unique one

            # https://www.elastic.co/guide/en/elasticsearch/reference/current/docs-bulk.html
            if op_type == "update":
                if "_id" not in action:
                    raise ValueError("Update requires _id")
                action["doc"] = doc
                action["doc_as_upsert"] = True
            else:
                action.update(doc)

            yield action

    actions = list(es_actions(index, documents, op_type))
    try:
        successes, failures = elasticsearch.helpers.bulk(
            es(),
            actions,
            stats_only=False,
            raise_on_error=raise_on_error,
        )
    except elasticsearch.helpers.BulkIndexError as e:
        logging.error("Error on indexing: " + json.dumps(e.errors, indent=2, default=str))
        if e.errors:
            _, error = list(e.errors[0].items())[0]
            reason = error.get("error", {}).get("reason", error)
            e.args = e.args + (f"First error: {reason}",)
        raise

    # Start preprocessors for this index (if any)
    processor.get_manager().start_index_preprocessors(index)

    return dict(successes=successes, failures=failures)


def get_document(index: str, doc_id: str, **kargs) -> dict:
    """
    Get a single document from this index.

    :param index: The name of the index
    :param doc_id: The document id (hash)
    :return: the source dict of the document
    """
    return es().get(index=index, id=doc_id, **kargs)["_source"]


def update_document(index: str, doc_id: str, fields: dict):
    """
    Update a single document.

    :param index: The name of the index
    :param doc_id: The document id (hash)
    :param fields: a {field: value} mapping of fields to update
    """
    # Mypy doesn't understand that body= has been deprecated already...
    es().update(index=index, id=doc_id, doc=fields)  # type: ignore


def delete_document(index: str, doc_id: str):
    """
    Delete a single document

    :param index: The Pname of the index
    :param doc_id: The document id (hash)
    """
    es().delete(index=index, id=doc_id)


UDATE_SCRIPTS = dict(
    add="""
    if (ctx._source[params.field] == null) {
      ctx._source[params.field] = [params.tag]
    } else {
      if (ctx._source[params.field].contains(params.tag)) {
        ctx.op = 'noop';
      } else {
        ctx._source[params.field].add(params.tag)
      }
    }
    """,
    remove="""
    if (ctx._source[params.field] != null && ctx._source[params.field].contains(params.tag)) {
      ctx._source[params.field].removeAll([params.tag]);
      if (ctx._source[params.field].size() == 0) {
        ctx._source.remove(params.field);
      }
    } else {
      ctx.op = 'noop';
    }
    """,
)


def update_tag_by_query(index: str | list[str], action: Literal["add", "remove"], query: dict, field: str, tag: str):
    create_or_verify_tag_field(index, field)
    script = dict(source=UDATE_SCRIPTS[action], lang="painless", params=dict(field=field, tag=tag))
    result = es().update_by_query(index=index, script=script, **query, refresh=True)
    return dict(updated=result["updated"], total=result["total"])


def update_documents_by_query(index: str | list[str], query: dict, field: str, value: Any):
    if value is None:
        script = dict(source="ctx._source.remove(params.field)", lang="painless", params=dict(field=field))
    else:
        script = dict(
            source="ctx._source[params.field] = params.value", lang="painless", params=dict(field=field, value=value)
        )
    return es().update_by_query(index=index, query=query, script=script, refresh=True)


### WvA Should probably move these to multimedia/actions or something


def get_instructions(index: str) -> Iterable[PreprocessingInstruction]:
    res = es().get(index=get_settings().system_index, id=index, source="preprocessing")
    for i in res["_source"].get("preprocessing", []):
        for a in i.get("arguments", []):
            if a.get("secret"):
                a["value"] = "********"
        yield PreprocessingInstruction.model_validate(i)


def get_instruction(index: str, field: str) -> Optional[PreprocessingInstruction]:
    for i in get_instructions(index):
        if i.field == field:
            return i


def add_instruction(index: str, instruction: PreprocessingInstruction):
    if instruction.field in get_fields(index):
        raise ValueError(f"Field {instruction.field} already exists in index {index}")
    instructions = list(get_instructions(index))
    instructions.append(instruction)
    create_fields(index, {instruction.field: "preprocess"})
    body = [i.model_dump() for i in instructions]
    es().update(index=get_settings().system_index, id=index, doc=dict(preprocessing=body))
    processor.get_manager().add_preprocessor(index, instruction)


def reassign_preprocessing_errors(index: str, field: str):
    """Reset status for any documents with error status, and restart preprocessor"""
    query = dict(term={f"{field}.status": dict(value="error")})
    update_documents_by_query(index, query, field, None)
    processor.get_manager().start_preprocessor(index, field)


def stop_preprocessor(index: str, field: str):
    processor.get_manager().stop_preprocessor(index, field)


def start_preprocessor(index: str, field: str):
    processor.get_manager().start_preprocessor(index, field)



"""API Endpoints for querying."""

from re import search
from typing import Annotated, Dict, List, Optional, Any, Union, Iterable, Literal

from fastapi import APIRouter, HTTPException, status, Depends, Response, Body
from pydantic import InstanceOf
from pydantic.main import BaseModel

from amcat4 import query, aggregate
from amcat4.aggregate import Axis, Aggregation
from amcat4.api.auth import authenticated_user, check_fields_access, check_role
from amcat4.config import AuthOptions, get_settings
from amcat4.fields import create_fields
from amcat4.index import Role, get_role, get_fields, update_documents_by_query
from amcat4.models import FieldSpec, FilterSpec, FilterValue, SortSpec
from amcat4.query import update_query, update_tag_query

app_query = APIRouter(prefix="/index", tags=["query"])


class QueryMeta(BaseModel):
    """Form for query metadata."""

    total_count: int
    per_page: Optional[int] = None
    page_count: Optional[int] = None
    page: Optional[int] = None
    scroll_id: Optional[str] = None


class QueryResult(BaseModel):
    """Form for query results."""

    results: List[Dict[str, Any]]
    meta: QueryMeta


def get_or_validate_allowed_fields(
    user: str, indices: Iterable[str], fields: list[FieldSpec] | None = None
) -> list[FieldSpec]:
    """
    For any endpoint that returns field values, make sure the user only gets fields that
    they are allowed to see. If fields is None, return all allowed fields. If fields is not None,
    check whether the user can access the fields (If not, raise an error).
    """
    if not isinstance(user, str):
        raise ValueError("User should be a string")
    if not isinstance(indices, list):
        raise ValueError("Indices should be a list")
    if fields is not None and not isinstance(fields, list):
        raise ValueError("Fields should be a list or None")

    no_auth = get_settings().auth == AuthOptions.no_auth
    if fields is None:
        if len(indices) > 1:
            # this restrictions is needed, because otherwise we need to return all allowed fields taking
            # into account the user's role for each index, and take the lowest possible access.
            # this is error prone and complex, so best to just disallow it. Also, requesting all fields
            # for multiple indices is probably not something we should support anyway
            raise ValueError("Fields should be specified if multiple indices are given")
        index_fields = get_fields(indices[0])
        role = get_role(indices[0], user)
        allowed_fields: list[FieldSpec] = []
        for field in index_fields.keys():
            if role >= Role.READER or no_auth:
                allowed_fields.append(FieldSpec(name=field))
            elif role == Role.METAREADER:
                metareader = index_fields[field].metareader
                if metareader.access == "read":
                    allowed_fields.append(FieldSpec(name=field))
                if metareader.access == "snippet":
                    allowed_fields.append(FieldSpec(name=field, snippet=metareader.max_snippet))
            else:
                raise HTTPException(
                    status_code=401,
                    detail=f"User {user} does not have a role on index {indices[0]}",
                )
        return allowed_fields

    for index in indices:
        if not no_auth:
            check_fields_access(index, user, fields)
    return fields


def _standardize_queries(queries: str | list[str] | dict[str, str] | None = None) -> dict[str, str] | None:
    """Convert query json to dict format: {label1:query1, label2: query2} uses indices if no labels given."""

    if queries:
        # to dict format: {label1:query1, label2: query2}  uses indices if no labels given
        if isinstance(queries, str):
            return {"1": queries}
        elif isinstance(queries, list):
            return {str(i): q for i, q in enumerate(queries)}
        elif isinstance(queries, dict):
            return queries
    return None


def _standardize_filters(
    filters: dict[str, FilterValue | list[FilterValue] | FilterSpec] | None = None
) -> dict[str, FilterSpec] | None:
    """Convert filters to dict format: {field: {values: []}}."""
    if not filters:
        return None

    f: dict[str, FilterSpec] = {}
    for field, filter_ in filters.items():
        if isinstance(filter_, str):
            f[field] = FilterSpec(values=[filter_])
        elif isinstance(filter_, list):
            f[field] = FilterSpec(values=filter_)
        elif isinstance(filter_, FilterSpec):
            f[field] = filter_
        else:
            raise ValueError(f"Cannot parse filter: {filter_}")
    return f


def _standardize_fieldspecs(fields: list[str | FieldSpec] | None = None) -> list[FieldSpec] | None:
    """Convert fields to list of FieldSpecs."""
    if not fields:
        return None

    f = []
    for field in fields:
        if isinstance(field, str):
            f.append(FieldSpec(name=field))
        elif isinstance(field, FieldSpec):
            f.append(field)
        else:
            raise ValueError(f"Cannot parse field: {field}")
    return f


def _standardize_sort(sort: str | list[str] | list[dict[str, SortSpec]] | None = None) -> list[dict[str, SortSpec]] | None:
    """Convert sort to list of dicts."""

    # TODO: sort cannot be right. that array around dict is useless

    if not sort:
        return None
    if isinstance(sort, str):
        return [{sort: SortSpec(order="asc")}]

    sortspec: list[dict[str, SortSpec]] = []

    for field in sort:
        if isinstance(field, str):
            sortspec.append({field: SortSpec(order="asc")})
        elif isinstance(field, dict):
            sortspec.append(field)
        else:
            raise ValueError(f"Cannot parse sort: {sort}")

    return sortspec


@app_query.post("/{index}/query", response_model=QueryResult)
def query_documents_post(
    index: str,
    queries: Annotated[
        str | list[str] | dict[str, str] | None,
        Body(
            description="Query/Queries to run. Value should be a single query string, a list of query strings, "
            "or a dict of {'label': 'query'}",
        ),
    ] = None,
    fields: Annotated[
        list[str | FieldSpec] | None,
        Body(
            description="List of fields to retrieve for each document"
            "In the list you can specify a fieldname, but also a FieldSpec dict."
            "Using the FieldSpec allows you to request only a snippet of a field."
            "fieldname[nomatch_chars;max_matches;match_chars]. 'matches' here refers to words from text queries. "
            "If there is no query, the snippet is the first [nomatch_chars] characters. "
            "If there is a query, snippets are returned for up to [max_matches] matches, with each match having [match_chars] "
            "characters. If there are multiple matches, they are concatenated with ' ... '.",
            openapi_examples={
                "simple": {"summary": "Retrieve single field", "value": '["title", "text", "date"]'},
                "text as snippet": {
                    "summary": "Retrieve the full title, but text only as snippet",
                    "value": '["title", {"name": "text", "snippet": {"nomatch_chars": 100}}]',
                },
                "all allowed fields": {
                    "summary": "If fields is left empty, all fields that the user is allowed to see are returned",
                },
            },
        ),
    ] = None,
    filters: Annotated[
        dict[str, FilterValue | list[FilterValue] | FilterSpec] | None,
        Body(
            description="Field filters, should be a dict of field names to filter specifications,"
            "which can be either a value, a list of values, or a FilterSpec dict",
        ),
    ] = None,
    sort: Annotated[
        str | list[str] | list[dict[str, SortSpec]] | None,
        Body(
            description="Sort by field name(s) or dict (see "
            "https://www.elastic.co/guide/en/elasticsearch/reference/current/sort-search-results.html for dict format)",
            openapi_examples={
                "simple": {"summary": "Sort by single field", "value": "'date'"},
                "multiple": {
                    "summary": "Sort by multiple fields",
                    "value": "['date', 'title']",
                },
                "dict": {
                    "summary": "Use dict to specify sort options",
                    "value": " [{'date': {'order':'desc'}}]",
                },
            },
        ),
    ] = None,
    per_page: Annotated[int, Body(le=200, description="Number of documents per page")] = 10,
    page: Annotated[int, Body(description="Which page to retrieve")] = 0,
    scroll: Annotated[
        str | None,
        Body(
            description="Scroll specification (e.g. '5m') to start a scroll request"
            "This will return a scroll_id which should be passed to subsequent calls"
            "(this is the advised way of scrolling through multiple pages of results)",
            examples=["5m"],
        ),
    ] = None,
    scroll_id: Annotated[str | None, Body(description="Scroll id from previous response to continue scrolling")] = None,
    highlight: Annotated[bool, Body(description="If true, highlight fields")] = False,
    user: str = Depends(authenticated_user),
):
    """
    List or query documents in this index.

    Returns a JSON object {data: [...], meta: {total_count, per_page, page_count, page|scroll_id}}
    """
    indices = index.split(",")
    fieldspecs = get_or_validate_allowed_fields(user, indices, _standardize_fieldspecs(fields))
    r = query.query_documents(
        indices,
        queries=_standardize_queries(queries),
        filters=_standardize_filters(filters),
        fields=fieldspecs,
        sort=_standardize_sort(sort),
        per_page=per_page,
        page=page,
        scroll_id=scroll_id,
        scroll=scroll,
        highlight=highlight,
    )
    if r is None:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail="No results")
    return r.as_dict()


class AggregationSpec(BaseModel):
    """Form for an aggregation."""

    # TODO: can probably merge wth Aggregation class?
    field: str
    function: str
    name: Optional[str] = None


class AxisSpec(BaseModel):
    """Form for an axis specification."""

    # TODO: can probably merge wth Axis class?
    field: str
    interval: Optional[str] = None


@app_query.post("/{index}/aggregate")
def query_aggregate_post(
    index: str,
    axes: Optional[List[AxisSpec]] = Body(None, description="Axes to aggregate on (i.e. group by)"),
    aggregations: Optional[List[AggregationSpec]] = Body(None, description="Aggregate functions to compute"),
    queries: Annotated[
        str | list[str] | dict[str, str] | None,
        Body(
            description="Query/Queries to run. Value should be a single query string, a list of query strings, "
            "or a dict of {'label': 'query'}",
        ),
    ] = None,
    filters: Annotated[
        dict[str, FilterValue | list[FilterValue] | FilterSpec] | None,
        Body(
            description="Field filters, should be a dict of field names to filter specifications,"
            "which can be either a value, a list of values, or a FilterSpec dict",
        ),
    ] = None,
    after: Annotated[dict[str, Any] | None, Body(description="After cursor for pagination")] = None,
    user: str = Depends(authenticated_user),
):
    """
    Construct an aggregate query.

    For example, to get average views per week per publisher
    {
     'axes': [{'field': 'date', 'interval':'week'}, {'field': 'publisher'}],
     'aggregations': [{'field': 'views', 'function': 'avg'}]
    }

    Returns a JSON object {data: [{axis1, ..., n, aggregate1, ...}, ...], meta: {axes: [...], aggregations: [...]}
    """

    indices = index.split(",")
    for index in indices:
        check_role(user, Role.READER, index)
    _axes = [Axis(**x.model_dump()) for x in axes] if axes else []
    _aggregations = [Aggregation(**x.model_dump()) for x in aggregations] if aggregations else []

    results = aggregate.query_aggregate(
        indices,
        _axes,
        _aggregations,
        queries=_standardize_queries(queries),
        filters=_standardize_filters(filters),
        after=after,
    )

    return {
        "meta": {
            "axes": [axis.asdict() for axis in results.axes],
            "aggregations": [a.asdict() for a in results.aggregations],
            "after": results.after,
        },
        "data": list(results.as_dicts()),
    }


@app_query.post("/{index}/tags_update")
def query_update_tags(
    index: str,
    action: Literal["add", "remove"] = Body(None, description="Action (add or remove) on tags"),
    field: str = Body(None, description="Tag field to update"),
    tag: str = Body(None, description="Tag to add or remove"),
    queries: Annotated[
        str | list[str] | dict[str, str] | None,
        Body(
            description="Query/Queries to run. Value should be a single query string, a list of query strings, "
            "or a dict of {'label': 'query'}",
        ),
    ] = None,
    filters: Annotated[
        dict[str, FilterValue | list[FilterValue] | FilterSpec] | None,
        Body(
            description="Field filters, should be a dict of field names to filter specifications,"
            "which can be either a value, a list of values, or a FilterSpec dict",
        ),
    ] = None,
    ids: Optional[Union[str, List[str]]] = Body(None, description="Document IDs of documents to update"),
    user: str = Depends(authenticated_user),
):
    """
    Add or remove tags by query or by id
    """
    indices = index.split(",")
    for i in indices:
        if get_role(i, user) < Role.WRITER:
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail=f"User {user} does not have permission to update tags on index {i}",
            )

    if isinstance(ids, (str, int)):
        ids = [ids]
    update_result = update_tag_query(
        indices, action, field, tag, _standardize_queries(queries), _standardize_filters(filters), ids
    )
    return update_result


@app_query.post("/{index}/update_by_query")
def update_by_query(
    index: str,
    field: Annotated[str, Body(description="Field to update")],
    value: Annotated[str | None, Body(description="New value for the field, or null/None to delete field")],
    queries: Annotated[
        str | list[str] | dict[str, str] | None,
        Body(
            description="Query/Queries to run. Value should be a single query string, a list of query strings, "
            "or a dict of {'label': 'query'}",
        ),
    ] = None,
    filters: Annotated[
        dict[str, FilterValue | list[FilterValue] | FilterSpec] | None,
        Body(
            description="Field filters, should be a dict of field names to filter specifications,"
            "which can be either a value, a list of values, or a FilterSpec dict",
        ),
    ] = None,
    user: str = Depends(authenticated_user),
):
    """
    Update documents by query.

    Select documents using queries and/or filters, and specify a field and new value.
    """
    check_role(user, Role.WRITER, index)
    return update_query(index, field, value, _standardize_queries(queries), _standardize_filters(filters))



"""
Multimedia features for AmCAT

AmCAT can link to a minio/S3 object store to provide access to multimedia content attached to documents.
The object store needs to be configured in the server settings.
"""

import datetime
from io import BytesIO
from multiprocessing import Value
from typing import Iterable, Optional
from venv import create
from amcat4.config import get_settings
from minio import Minio, S3Error
from minio.deleteobjects import DeleteObject
from minio.datatypes import PostPolicy, Object
import functools


def get_minio() -> Minio:
    result = connect_minio()
    if result is None:
        raise ValueError("Could not connect to minio")
    return result


@functools.lru_cache()
def connect_minio() -> Optional[Minio]:
    try:
        return _connect_minio()
    except Exception as e:
        raise Exception(f"Cannot connect to minio {get_settings().minio_host!r}: {e}")


def _connect_minio() -> Optional[Minio]:
    settings = get_settings()
    if settings.minio_host is None:
        return None
    if settings.minio_secret_key is None or settings.minio_access_key is None:
        raise ValueError("minio_access_key or minio_secret_key not specified")
    return Minio(
        settings.minio_host,
        secure=settings.minio_tls,
        access_key=settings.minio_access_key,
        secret_key=settings.minio_secret_key,
    )


def bucket_name(index: str) -> str:
    return index.replace("_", "-")


def get_bucket(minio: Minio, index: str, create_if_needed=True):
    """
    Get the bucket name for this index. If create_if_needed is True, create the bucket if it doesn't exist.
    Returns the bucket name, or "" if it doesn't exist and create_if_needed is False.
    """
    bucket = bucket_name(index)
    if not minio.bucket_exists(bucket):
        if not create_if_needed:
            return ""
        minio.make_bucket(bucket)
    return bucket


def list_multimedia_objects(
    index: str, prefix: Optional[str] = None, start_after: Optional[str] = None, recursive=True
) -> Iterable[Object]:
    minio = get_minio()
    bucket = get_bucket(minio, index, create_if_needed=False)
    if not bucket:
        return
    yield from minio.list_objects(bucket_name(index), prefix=prefix, start_after=start_after, recursive=recursive)


def stat_multimedia_object(index: str, key: str) -> Object:
    minio = get_minio()
    bucket = get_bucket(minio, index, create_if_needed=False)
    if not bucket:
        raise ValueError(f"Bucket for {index} does not exist")
    return minio.stat_object(bucket, key)


def get_multimedia_object(index: str, key: str) -> bytes:
    minio = get_minio()
    bucket = get_bucket(minio, index, create_if_needed=False)
    if not bucket:
        raise ValueError(f"Bucket for {index} does not exist")
    res = minio.get_object(bucket, key)
    return res.read()


def delete_bucket(minio: Minio, index: str):
    bucket = get_bucket(minio, index, create_if_needed=False)
    if not bucket:
        return
    to_delete = [DeleteObject(x.object_name) for x in minio.list_objects(bucket, recursive=True) if x.object_name]
    errors = list(minio.remove_objects(bucket, to_delete))
    if errors:
        raise Exception(f"Error on deleting objects: {errors}")
    minio.remove_bucket(bucket)


def add_multimedia_object(index: str, key: str, bytes: bytes):
    minio = get_minio()
    bucket = get_bucket(minio, index)
    data = BytesIO(bytes)
    minio.put_object(bucket, key, data, len(bytes))


def presigned_post(index: str, key_prefix: str = "", days_valid=1):
    minio = get_minio()
    bucket = get_bucket(minio, index)
    policy = PostPolicy(bucket, expiration=datetime.datetime.now() + datetime.timedelta(days=days_valid))
    policy.add_starts_with_condition("key", key_prefix)
    url = f"http{'s' if get_settings().minio_tls else ''}://{get_settings().minio_host}/{bucket}"
    return url, minio.presigned_post_policy(policy)


def presigned_get(index: str, key, days_valid=1):
    minio = get_minio()
    bucket = get_bucket(minio, index)
    return minio.presigned_get_object(bucket, key, expires=datetime.timedelta(days=days_valid))



"""
We have two types of fields:
- Elastic fields are the fields used under the hood by elastic.
  (https://www.elastic.co/guide/en/elasticsearch/reference/current/mapping-types.html
  These are stored in the Mapping of an index
- Amcat fields (Field) are the fields are seen by the amcat user. They use a simplified type, and contain additional
  information such as metareader access
  These are stored in the system index.

We need to make sure that:
- When a user sets a field, it needs to be changed in both types: the system index and the mapping
- If a field only exists in the elastic mapping, we need to add the default Field to the system index.
  This happens anytime get_fields is called, so that whenever a field is used it is guarenteed to be in the
  system index
"""

import datetime
import json
from typing import Any, Iterator, Literal, Mapping, get_args, cast


from elasticsearch import NotFoundError

# from amcat4.api.common import py2dict
from amcat4.config import get_settings
from amcat4.elastic import es
from amcat4.models import FieldType, CreateField, ElasticType, Field, UpdateField, FieldMetareaderAccess


# given an elastic field type, Check if it is supported by AmCAT.
# this is not just the inverse of TYPEMAP_AMCAT_TO_ES because some AmCAT types map to multiple elastic
# types (e.g., tag and keyword, image_url and wildcard)
# (this is relevant if we are importing an index)
TYPEMAP_ES_TO_AMCAT: dict[ElasticType, FieldType] = {
    # TEXT fields
    "text": "text",
    "annotated_text": "text",
    "binary": "text",
    "match_only_text": "text",
    # DATE fields
    "date": "date",
    # BOOLEAN fields
    "boolean": "boolean",
    # KEYWORD fields
    "keyword": "keyword",
    "constant_keyword": "keyword",
    "wildcard": "keyword",
    # INTEGER fields
    "integer": "number",
    "byte": "number",
    "short": "number",
    "long": "number",
    "unsigned_long": "number",
    # NUMBER fields
    "float": "number",
    "half_float": "number",
    "double": "number",
    "scaled_float": "number",
    # OBJECT fields
    "object": "object",
    "flattened": "object",
    "nested": "object",
    # VECTOR fields (exclude sparse vectors)
    "dense_vector": "vector",
    # GEO fields
    "geo_point": "geo_point",
}

# maps amcat field types to elastic field types.
# The first elastic type in the array is the default.
TYPEMAP_AMCAT_TO_ES: dict[FieldType, list[ElasticType]] = {
    "text": ["text", "annotated_text", "binary", "match_only_text"],
    "date": ["date"],
    "boolean": ["boolean"],
    "keyword": ["keyword", "constant_keyword", "wildcard"],
    "number": ["double", "float", "half_float", "scaled_float"],
    "integer": ["long", "integer", "byte", "short", "unsigned_long"],
    "object": ["object", "flattened", "nested"],
    "vector": ["dense_vector"],
    "geo_point": ["geo_point"],
    "tag": ["keyword", "wildcard"],
    "image": ["wildcard", "keyword", "constant_keyword", "text"],
    "video": ["wildcard", "keyword", "constant_keyword", "text"],
    "audio": ["wildcard", "keyword", "constant_keyword", "text"],
    "url": ["wildcard", "keyword", "constant_keyword", "text"],
    "json": ["text"],
    "preprocess": ["object"],
}


def get_default_metareader(type: FieldType):
    if type in ["boolean", "number", "date"]:
        return FieldMetareaderAccess(access="read")

    return FieldMetareaderAccess(access="none")


def get_default_field(type: FieldType):
    """
    Generate a field on the spot with default settings.
    Primary use case is importing existing indices with fields that are not registered in the system index.
    """
    elastic_type = TYPEMAP_AMCAT_TO_ES.get(type)
    if elastic_type is None:
        raise ValueError(
            f"The default elastic type mapping for field type {type} is not defined (if this happens, blame and inform Kasper)"
        )
    return Field(elastic_type=elastic_type[0], type=type, metareader=get_default_metareader(type))


def _standardize_createfields(fields: Mapping[str, FieldType | CreateField]) -> dict[str, CreateField]:
    sfields = {}
    for k, v in fields.items():
        if isinstance(v, str):
            assert v in get_args(FieldType), f"Unknown amcat type {v}"
            sfields[k] = CreateField(type=cast(FieldType, v))
        else:
            sfields[k] = v
    return sfields


def check_forbidden_type(field: Field, type: FieldType):
    if field.identifier:
        for forbidden_type in ["tag", "vector"]:
            if type == forbidden_type:
                raise ValueError(f"Field {field} is an identifier field, which cannot be a {forbidden_type} field")


def coerce_type(value: Any, type: FieldType):
    """
    Coerces values into the respective type in elastic
    based on ES_MAPPINGS and elastic field types
    """
    if type == "date" and isinstance(value, datetime.date):
        return value.isoformat()
    if type in ["text", "tag", "image", "video", "audio", "date"]:
        return str(value)
    if type in ["boolean"]:
        return bool(value)
    if type in ["number"]:
        return float(value)
    if type in ["integer"]:
        return int(value)

    if type == "json":
        if isinstance(value, str):
            return value
        return json.dumps(value)

    # TODO: check coercion / validation for object, vector and geo types
    if type in ["object"]:
        return value
    if type in ["vector"]:
        return value
    if type in ["geo_point"]:
        return value

    return value


def create_fields(index: str, fields: Mapping[str, FieldType | CreateField]):
    mapping: dict[str, Any] = {}
    current_fields = get_fields(index)

    sfields = _standardize_createfields(fields)
    old_identifiers = any(f.identifier for f in current_fields.values())
    new_identifiers = False

    for field, settings in sfields.items():
        if settings.elastic_type is not None:
            allowed_types = TYPEMAP_AMCAT_TO_ES.get(settings.type, [])
            if settings.elastic_type not in allowed_types:
                raise ValueError(
                    f"Field type {settings.type} does not support elastic type {settings.elastic_type}. "
                    f"Allowed types are: {allowed_types}"
                )
            elastic_type = settings.elastic_type
        else:
            elastic_type = get_default_field(settings.type).elastic_type

        current = current_fields.get(field)
        if current is not None:
            # fields can already exist. For example, a scraper might include the field types in every
            # upload request. If a field already exists, we'll ignore it, but we will throw an error
            # if static settings (elastic type, identifier) do not match.
            if current.elastic_type != elastic_type:
                raise ValueError(f"Field '{field}' already exists with elastic type '{current.elastic_type}'. ")
            if current.identifier != settings.identifier:
                raise ValueError(f"Field '{field}' already exists with identifier '{current.identifier}'. ")
            continue

        # if field does not exist, we add it to both the mapping and the system index
        if settings.identifier:
            new_identifiers = True
        mapping[field] = {"type": elastic_type}
        if settings.type == "preprocess":
            mapping[field]["properties"] = dict(status=dict(type="keyword"), error=dict(type="text", index=False))
        if settings.type in ["date"]:
            mapping[field]["format"] = "strict_date_optional_time"

        current_fields[field] = Field(
            type=settings.type,
            elastic_type=elastic_type,
            identifier=settings.identifier,
            metareader=settings.metareader or get_default_metareader(settings.type),
            client_settings=settings.client_settings or {},
        )
        check_forbidden_type(current_fields[field], settings.type)

    if new_identifiers:
        # new identifiers are only allowed if the index had identifiers, or if it is a new index (i.e. no documents)
        has_docs = es().count(index=index)["count"] > 0
        if has_docs and not old_identifiers:
            raise ValueError("Cannot add identifiers. Index already has documents with no identifiers.")

    if len(mapping) > 0:
        # if there are new identifiers, check whether this is allowed first
        es().indices.put_mapping(index=index, properties=mapping)
        es().update(
            index=get_settings().system_index,
            id=index,
            doc=dict(fields=_fields_to_elastic(current_fields)),
        )


def _fields_to_elastic(fields: dict[str, Field]) -> list[dict]:
    # some additional validation
    return [{"field": field, "settings": settings.model_dump()} for field, settings in fields.items()]


def _fields_from_elastic(
    fields: list[dict],
) -> dict[str, Field]:
    return {fs["field"]: Field.model_validate(fs["settings"]) for fs in fields}


def update_fields(index: str, fields: dict[str, UpdateField]):
    """
    Set the fields settings for this index. Only updates fields that
    already exist. Only keys in UpdateField can be updated (not type or client_settings)
    """

    current_fields = get_fields(index)

    for field, new_settings in fields.items():
        current = current_fields.get(field)
        if current is None:
            raise ValueError(f"Field {field} does not exist")

        if new_settings.type is not None:
            check_forbidden_type(current, new_settings.type)

            valid_es_types = TYPEMAP_AMCAT_TO_ES.get(new_settings.type)
            if valid_es_types is None:
                raise ValueError(f"Invalid field type: {new_settings.type}")
            if current.elastic_type not in valid_es_types:
                raise ValueError(
                    f"Field {field} has the elastic type {current.elastic_type}. A {new_settings.type} field can only have the following elastic types: {valid_es_types}."
                )
            current_fields[field].type = new_settings.type

        if new_settings.metareader is not None:
            if current.type != "text" and new_settings.metareader.access == "snippet":
                raise ValueError(f"Field {field} is not of type text, cannot set metareader access to snippet")
            current_fields[field].metareader = new_settings.metareader

        if new_settings.client_settings is not None:
            current_fields[field].client_settings = new_settings.client_settings

    es().update(
        index=get_settings().system_index,
        id=index,
        doc=dict(fields=_fields_to_elastic(current_fields)),
    )


def _get_index_fields(index: str) -> Iterator[tuple[str, ElasticType]]:
    r = es().indices.get_mapping(index=index)

    if len(r[index]["mappings"]) > 0:
        for k, v in r[index]["mappings"]["properties"].items():
            yield k, v.get("type", "object")


def get_fields(index: str) -> dict[str, Field]:
    """
    Retrieve the fields settings for this index. Look for both the field settings in the system index,
    and the field mappings in the index itself. If a field is not defined in the system index, return the
    default settings for that field type and add it to the system index. This way, any elastic index can be imported
    """
    fields: dict[str, Field] = {}
    system_index = get_settings().system_index

    try:
        d = es().get(
            index=system_index,
            id=index,
            source_includes="fields",
        )
        system_index_fields = _fields_from_elastic(d["_source"].get("fields", {}))
    except NotFoundError:
        system_index_fields = {}

    update_system_index = False
    for field, elastic_type in _get_index_fields(index):
        type = TYPEMAP_ES_TO_AMCAT.get(elastic_type)

        if type is None:
            # skip over unsupported elastic fields.
            # (TODO: also return warning to client?)
            continue

        if field not in system_index_fields:
            update_system_index = True
            fields[field] = get_default_field(type)
        else:
            fields[field] = system_index_fields[field]

    if update_system_index:
        es().update(
            index=system_index,
            id=index,
            doc=dict(fields=_fields_to_elastic(fields)),
        )

    return fields


def create_or_verify_tag_field(index: str | list[str], field: str):
    """Create a special type of field that can be used to tag documents.
    Since adding/removing tags supports multiple indices, we first check whether the field name is valid for all indices"""
    indices = [index] if isinstance(index, str) else index
    add_to_indices = []
    for i in indices:
        current_fields = get_fields(i)
        if field in current_fields:
            if current_fields[field].type != "tag":
                raise ValueError(f"Field '{field}' already exists in index '{i}' and is not a tag field")

    else:
        add_to_indices.append(i)

    for i in add_to_indices:
        current_fields[field] = get_default_field("tag")
        es().indices.put_mapping(index=index, properties={field: {"type": "keyword"}})
        es().update(
            index=get_settings().system_index,
            id=i,
            doc=dict(fields=_fields_to_elastic(current_fields)),
        )


def field_values(index: str, field: str, size: int) -> list[str]:
    """
    Get the values for a given field (e.g. to populate list of filter values on keyword field)
    Results are sorted descending by document frequency
    see: https://www.elastic.co/guide/en/elasticsearch/reference/7.4/search-aggregations-bucket-terms-aggregation.html
         #search-aggregations-bucket-terms-aggregation-order

    :param index: The index
    :param field: The field name
    :return: A list of values
    """
    aggs = {"unique_values": {"terms": {"field": field, "size": size}}}
    r = es().search(index=index, size=0, aggs=aggs)
    return [x["key"] for x in r["aggregations"]["unique_values"]["buckets"]]


def field_stats(index: str, field: str) -> list[str]:
    """
    :param index: The index
    :param field: The field name
    :return: A list of values
    """
    aggs = {"facets": {"stats": {"field": field}}}
    r = es().search(index=index, size=0, aggs=aggs)
    return r["aggregations"]["facets"]


def update_by_query(index: str | list[str], script: str, query: dict, params: dict | None = None):
    script_dict = dict(source=script, lang="painless", params=params or {})
    es().update_by_query(index=index, script=script_dict, **query)



"""API Endpoints for document and index management."""

from http import HTTPStatus
from re import U
from typing import Annotated, Any, Literal

import elasticsearch
from elastic_transport import ApiError
from fastapi import APIRouter, HTTPException, Response, status, Depends, Body
from pydantic import BaseModel
from datetime import datetime

from amcat4 import index, fields as index_fields
from amcat4.api.auth import authenticated_user, authenticated_writer, check_role

from amcat4.index import refresh_system_index, remove_role, set_role
from amcat4.fields import field_values, field_stats
from amcat4.models import CreateField, FieldType, UpdateField

app_index = APIRouter(prefix="/index", tags=["index"])

RoleType = Literal["ADMIN", "WRITER", "READER", "METAREADER"]


@app_index.get("/")
def index_list(current_user: str = Depends(authenticated_user)):
    """
    List index from this server.

    Returns a list of dicts containing name, role, and guest attributes
    """

    def index_to_dict(ix: index.Index) -> dict:
        ix_dict = ix._asdict()
        guest_role_int = ix_dict.get("guest_role", 0)

        ix_dict = dict(
            id=ix_dict["id"],
            name=ix_dict["name"],
            guest_role=index.Role(guest_role_int).name,
            description=ix_dict.get("description", ""),
            archived=ix_dict.get("archived", ""),
        )
        return ix_dict

    return [index_to_dict(ix) for ix in index.list_known_indices(current_user)]


class NewIndex(BaseModel):
    """Form to create a new index."""

    id: str
    name: str | None = None
    guest_role: RoleType | None = None
    description: str | None = None


@app_index.post("/", status_code=status.HTTP_201_CREATED)
def create_index(new_index: NewIndex, current_user: str = Depends(authenticated_writer)):
    """
    Create a new index, setting the current user to admin (owner).

    POST data should be json containing name and optional guest_role
    """
    guest_role = new_index.guest_role and index.Role[new_index.guest_role.upper()]
    try:
        index.create_index(
            new_index.id,
            guest_role=guest_role,
            name=new_index.name,
            description=new_index.description,
            admin=current_user if current_user != "_admin" else None,
        )
    except ApiError as e:
        raise HTTPException(
            status_code=400,
            detail=dict(info=f"Error on creating index: {e}", message=e.message, body=e.body),
        )


# TODO Yes, this should be linked to the actual roles enum
class ChangeIndex(BaseModel):
    """Form to update an existing index."""

    name: str | None = None
    description: str | None = None
    guest_role: Literal["WRITER", "READER", "METAREADER", "NONE"] | None = None
    archive: bool | None = None


@app_index.put("/{ix}")
def modify_index(ix: str, data: ChangeIndex, user: str = Depends(authenticated_user)):
    """
    Modify the index.

    POST data should be json containing the changed values (i.e. name, description, guest_role)

    User needs admin rights on the index
    """
    check_role(user, index.Role.ADMIN, ix)
    guest_role = index.GuestRole[data.guest_role] if data.guest_role is not None else None
    archived = None
    if data.archive is not None:
        d = index.get_index(ix)
        is_archived = d.archived is not None and d.archived != ""
        if is_archived != data.archive:
            archived = str(datetime.now()) if data.archive else ""

    index.modify_index(
        ix,
        name=data.name,
        description=data.description,
        guest_role=guest_role,
        archived=archived,
        # remove_guest_role=remove_guest_role,
        # unarchive=unarchive,
    )
    refresh_system_index()


@app_index.get("/{ix}")
def view_index(ix: str, user: str = Depends(authenticated_user)):
    """
    View the index.
    """
    try:
        role = check_role(user, index.Role.METAREADER, ix, required_global_role=index.Role.WRITER)
        d = index.get_index(ix)._asdict()
        d["user_role"] = role.name
        d["guest_role"] = index.GuestRole(d.get("guest_role", 0)).name
        d["description"] = d.get("description", "") or ""
        d["name"] = d.get("name", "") or ""
        return d
    except index.IndexDoesNotExist:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=f"Index {ix} does not exist")


@app_index.post("/{ix}/archive", status_code=status.HTTP_204_NO_CONTENT, response_class=Response)
def archive_index(
    ix: str,
    archived: Annotated[bool, Body(description="Boolean for setting archived to true or false")],
    user: str = Depends(authenticated_user),
):
    """Archive or unarchive the index. When an index is archived, it restricts usage, and adds a timestamp for when
    it was archived. An index can only be deleted if it has been archived for a specific amount of time."""
    check_role(user, index.Role.ADMIN, ix)
    try:
        d = index.get_index(ix)
        is_archived = d.archived is not None
        if is_archived == archived:
            return
        archived_date = str(datetime.now()) if archived else None
        index.modify_index(ix, archived=archived_date)

    except index.IndexDoesNotExist:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=f"Index {ix} does not exist")


@app_index.delete("/{ix}", status_code=status.HTTP_204_NO_CONTENT, response_class=Response)
def delete_index(ix: str, user: str = Depends(authenticated_user)):
    """Delete the index."""
    check_role(user, index.Role.ADMIN, ix)
    try:
        index.delete_index(ix)
    except index.IndexDoesNotExist:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=f"Index {ix} does not exist")


@app_index.post("/{ix}/documents", status_code=status.HTTP_201_CREATED)
def upload_documents(
    ix: str,
    documents: Annotated[list[dict[str, Any]], Body(description="The documents to upload")],
    fields: Annotated[
        dict[str, FieldType | CreateField] | None,
        Body(
            description="If a field in documents does not yet exist, you can create it on the spot. "
            "If you only need to specify the type, and use the default settings, "
            "you can use the short form: {field: type}"
        ),
    ] = None,
    operation: Annotated[
        Literal["update", "create"],
        Body(
            description="The operation to perform. Default is create, which ignores any documents that already exist. "
            "The 'update' operation behaves as an upsert (create or update). If an identical document (or document with "
            "identical identifiers) already exists, the uploaded fields will be created or overwritten. If there are fields "
            "in the original document that are not in the uploaded document, they will NOT be removed. since update is destructive "
            "it requires admin rights."
        ),
    ] = "create",
    user: str = Depends(authenticated_user),
):
    """
    Upload documents to this server. Returns a list of ids for the uploaded documents
    """
    if operation == "create":
        check_role(user, index.Role.WRITER, ix)
    else:
        check_role(user, index.Role.ADMIN, ix)
    return index.upload_documents(ix, documents, fields, operation)


@app_index.get("/{ix}/documents/{docid}")
def get_document(
    ix: str,
    docid: str,
    fields: str | None = None,
    user: str = Depends(authenticated_user),
):
    """
    Get a single document by id.

    GET request parameters:
    fields - Comma separated list of fields to return (default: all fields)
    """
    check_role(user, index.Role.READER, ix)
    kargs = {}
    if fields:
        kargs["_source"] = fields
    try:
        return index.get_document(ix, docid, **kargs)
    except elasticsearch.exceptions.NotFoundError:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"Document {ix}/{docid} not found",
        )


@app_index.put(
    "/{ix}/documents/{docid}",
    status_code=status.HTTP_204_NO_CONTENT,
    response_class=Response,
)
def update_document(
    ix: str,
    docid: str,
    update: dict = Body(...),
    user: str = Depends(authenticated_user),
):
    """
    Update a document.

    PUT request body should be a json {field: value} mapping of fields to update
    """
    check_role(user, index.Role.WRITER, ix)
    try:
        index.update_document(ix, docid, update)
    except elasticsearch.exceptions.NotFoundError:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"Document {ix}/{docid} not found",
        )


@app_index.delete(
    "/{ix}/documents/{docid}",
    status_code=status.HTTP_204_NO_CONTENT,
    response_class=Response,
)
def delete_document(ix: str, docid: str, user: str = Depends(authenticated_user)):
    """Delete this document."""
    check_role(user, index.Role.WRITER, ix)
    try:
        index.delete_document(ix, docid)
    except elasticsearch.exceptions.NotFoundError:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"Document {ix}/{docid} not found",
        )


@app_index.post("/{ix}/fields")
def create_fields(
    ix: str,
    fields: Annotated[
        dict[str, FieldType | CreateField],
        Body(
            description="Either a dictionary that maps field names to field specifications"
            "({field: {type: 'text', identifier: True }}), "
            "or a simplified version that only specifies the type ({field: type})"
        ),
    ],
    user: str = Depends(authenticated_user),
):
    """
    Create fields
    """
    check_role(user, index.Role.WRITER, ix)
    index_fields.create_fields(ix, fields)
    return "", HTTPStatus.NO_CONTENT


@app_index.get("/{ix}/fields")
def get_fields(ix: str, user: str = Depends(authenticated_user)):
    """
    Get the fields (columns) used in this index.

    Returns a json array of {name, type} objects
    """
    check_role(user, index.Role.METAREADER, ix)
    return index.get_fields(ix)


@app_index.put("/{ix}/fields")
def update_fields(
    ix: str, fields: Annotated[dict[str, UpdateField], Body(description="")], user: str = Depends(authenticated_user)
):
    """
    Update the field settings
    """
    check_role(user, index.Role.WRITER, ix)

    index_fields.update_fields(ix, fields)
    return "", HTTPStatus.NO_CONTENT


@app_index.get("/{ix}/fields/{field}/values")
def get_field_values(ix: str, field: str, user: str = Depends(authenticated_user)):
    """
    Get unique values for a specific field. Should mainly/only be used for tag fields.
    Main purpose is to provide a list of values for a dropdown menu.

    TODO: at the moment 'only' returns top 2000 values. Currently throws an
    error if there are more than 2000 unique values. We can increase this limit, but
    there should be a limit. Querying could be an option, but not sure if that is
    efficient, since elastic has to aggregate all values first.
    """
    check_role(user, index.Role.READER, ix)
    values = field_values(ix, field, size=2001)
    if len(values) > 2000:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=f"Field {field} has more than 2000 unique values",
        )
    return values


@app_index.get("/{ix}/fields/{field}/stats")
def get_field_stats(ix: str, field: str, user: str = Depends(authenticated_user)):
    """Get statistics for a specific value. Only works for numeric (incl date) fields."""
    check_role(user, index.Role.READER, ix)
    return field_stats(ix, field)


@app_index.get("/{ix}/users")
def list_index_users(ix: str, user: str = Depends(authenticated_user)):
    """
    List the users in this index.

    Allowed for global admin and local readers
    """
    if index.get_global_role(user) != index.Role.ADMIN:
        check_role(user, index.Role.READER, ix)
    return [{"email": u, "role": r.name} for (u, r) in index.list_users(ix).items()]


def _check_can_modify_user(ix, user, target_user, target_role):
    if index.get_global_role(user) != index.Role.ADMIN:
        required_role = (
            index.Role.ADMIN
            if (target_role == index.Role.ADMIN or index.get_role(ix, target_user) == index.Role.ADMIN)
            else index.Role.WRITER
        )
        check_role(user, required_role, ix)


@app_index.post("/{ix}/users", status_code=status.HTTP_201_CREATED)
def add_index_users(
    ix: str,
    email: str = Body(..., description="Email address of the user to add"),
    role: RoleType = Body(..., description="Role of the user to add"),
    user: str = Depends(authenticated_user),
):
    """
    Add an existing user to this index.

    To create regular users you need WRITER permission. To create ADMIN users, you need ADMIN permission.
    Global ADMINs can always add users.
    """
    r = index.Role[role]
    _check_can_modify_user(ix, user, email, r)
    set_role(ix, email, r)
    return {"user": email, "index": ix, "role": r.name}


@app_index.put("/{ix}/users/{email}")
def modify_index_user(
    ix: str,
    email: str,
    role: RoleType = Body(..., description="New role for the user", embed=True),
    user: str = Depends(authenticated_user),
):
    """
    Change the role of an existing user.

    This requires WRITER rights on the index or global ADMIN rights.
    If changing a user from or to ADMIN, it requires (local or global) ADMIN rights
    """
    r = index.Role[role]
    _check_can_modify_user(ix, user, email, r)
    set_role(ix, email, r)
    return {"user": email, "index": ix, "role": r.name}


@app_index.delete("/{ix}/users/{email}")
def remove_index_user(ix: str, email: str, user: str = Depends(authenticated_user)):
    """
    Remove this user from the index.

    This requires WRITER rights on the index.
    If removing an ADMIN user, it requires ADMIN rights
    """
    _check_can_modify_user(ix, user, email, None)
    remove_role(ix, email)
    return {"user": email, "index": ix, "role": None}


@app_index.get("/{ix}/refresh", status_code=status.HTTP_204_NO_CONTENT, response_class=Response)
def refresh_index(ix: str):
    index.refresh_index(ix)



from inspect import isclass
from typing import Optional, Iterable


class DateMapping:
    interval: str | None = None

    def mapping(self, field: str) -> dict:
        return {self.fieldname(field): {"type": self.mapping_type(), "script": self.mapping_script(field)}}

    def mapping_script(self, field: str) -> str:
        raise NotImplementedError()

    def mapping_type(self):
        return "keyword"

    def fieldname(self, field: str) -> str:
        return f"{field}_{self.interval}"

    def postprocess(self, value):
        return value


class DayOfWeek(DateMapping):
    interval = "dayofweek"

    def mapping_script(self, field):
        return f"emit(doc['{field}'].value.dayOfWeekEnum.getDisplayName(TextStyle.FULL, Locale.ROOT))"


class DayPart(DateMapping):
    interval = "daypart"

    def mapping_script(self, field):
        return """
            int hour =doc['date'].value.hour;
            if (hour < 6) emit('Night');
            else if (hour < 12) emit('Morning');
            else if (hour < 18) emit('Afternoon');
            else emit('Evening')
        """


class MonthName(DateMapping):
    interval = "monthnr"

    def mapping_type(self):
        return "double"

    def mapping_script(self, field):
        return "emit(doc['date'].value.getMonthValue())"

    def postprocess(self, value):
        return int(value)


class YearNr(DateMapping):
    interval = "yearnr"

    def mapping_type(self):
        return "double"

    def mapping_script(self, field):
        return "emit(doc['date'].value.getYear())"

    def postprocess(self, value):
        return int(value)


class DayOfMonth(DateMapping):
    interval = "dayofmonth"

    def mapping_type(self):
        return "double"

    def mapping_script(self, field):
        return "emit(doc['date'].value.getDayOfMonth())"

    def postprocess(self, value):
        return int(value)


class Weeknr(DateMapping):
    interval = "weeknr"

    def mapping_type(self):
        return "double"

    def mapping_script(self, field):
        return "emit(doc['date'].value.get(IsoFields.WEEK_OF_WEEK_BASED_YEAR))"

    def postprocess(self, value):
        return int(value)


def interval_mapping(interval: str | None) -> Optional[DateMapping]:
    if interval is not None:
        for m in mappings():
            if m.interval == interval:
                return m
    return None


def mappings() -> Iterable[DateMapping]:
    for c in globals().values():
        if isclass(c) and issubclass(c, DateMapping) and c != DateMapping and c.interval is not None:
            yield c()



import asyncio
import logging
from typing import Annotated, Literal
from fastapi import APIRouter, Body, Depends, HTTPException, Response, status

from amcat4 import index
from amcat4.api.auth import authenticated_user, check_role
from amcat4.preprocessing.models import PreprocessingInstruction
from amcat4.index import (
    get_instruction,
    get_instructions,
    add_instruction,
    reassign_preprocessing_errors,
    start_preprocessor,
    stop_preprocessor,
)
from amcat4.preprocessing.processor import get_counts, get_manager
from amcat4.preprocessing.task import get_tasks

logger = logging.getLogger("amcat4.preprocessing")

app_preprocessing = APIRouter(tags=["preprocessing"])


@app_preprocessing.get("/preprocessing_tasks")
def list_tasks():
    return [t.model_dump() for t in get_tasks()]


@app_preprocessing.get("/index/{ix}/preprocessing")
def list_instructions(ix: str, user: str = Depends(authenticated_user)):
    check_role(user, index.Role.READER, ix)
    return get_instructions(ix)


@app_preprocessing.post("/index/{ix}/preprocessing", status_code=status.HTTP_204_NO_CONTENT, response_class=Response)
async def post_instruction(ix: str, instruction: PreprocessingInstruction, user: str = Depends(authenticated_user)):
    check_role(user, index.Role.WRITER, ix)
    add_instruction(ix, instruction)


@app_preprocessing.get("/index/{ix}/preprocessing/{field}")
async def get_instruction_details(ix: str, field: str, user: str = Depends(authenticated_user)):
    check_role(user, index.Role.WRITER, ix)
    i = get_instruction(ix, field)
    if i is None:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"Preprocessing instruction for field {field} on index {ix} not found",
        )
    state = get_manager().get_status(ix, field)
    counts = get_counts(ix, field)
    return dict(instruction=i, status=state, counts=counts)


@app_preprocessing.get("/index/{ix}/preprocessing/{field}/status")
async def get_status(ix: str, field: str, user: str = Depends(authenticated_user)):
    return dict(status=get_manager().get_status(ix, field))


@app_preprocessing.post(
    "/index/{ix}/preprocessing/{field}/status", status_code=status.HTTP_204_NO_CONTENT, response_class=Response
)
async def set_status(
    ix: str,
    field: str,
    user: str = Depends(authenticated_user),
    action: Literal["Start", "Stop", "Reassign"] = Body(description="Status to set for this preprocessing task", embed=True),
):
    check_role(user, index.Role.WRITER, ix)
    current_status = get_manager().get_status(ix, field)
    if action == "Start" and current_status in {"Unknown", "Error", "Stopped", "Done"}:
        start_preprocessor(ix, field)
    elif action == "Stop" and current_status in {"Active"}:
        stop_preprocessor(ix, field)
    elif action == "Reassign":
        reassign_preprocessing_errors(ix, field)
    else:
        raise HTTPException(422, f"Cannot {action}, (status: {current_status}; field {ix}.{field})")



import asyncio
from functools import cache
import logging
from typing import Dict, Literal, Tuple
from elasticsearch import NotFoundError
from httpx import AsyncClient, HTTPStatusError
from amcat4.elastic import es
import amcat4.index
from amcat4.preprocessing.models import PreprocessingInstruction

logger = logging.getLogger("amcat4.preprocessing")

PreprocessorStatus = Literal["Active", "Paused", "Unknown", "Error", "Stopped", "Done"]


class RateLimit(Exception):
    pass


PAUSE_ON_RATE_LIMIT_SECONDS = 10


class PreprocessorManager:
    SINGLETON = None

    def __init__(self):
        self.preprocessors: Dict[Tuple[str, str], PreprocessingInstruction] = {}
        self.running_tasks: Dict[Tuple[str, str], asyncio.Task] = {}
        self.preprocessor_status: Dict[Tuple[str, str], PreprocessorStatus] = {}

    def set_status(self, index: str, field: str, status: PreprocessorStatus):
        self.preprocessor_status[index, field] = status

    def add_preprocessor(self, index: str, instruction: PreprocessingInstruction):
        """Start a new preprocessor task and add it to the manager, returning the Task object"""
        self.preprocessors[index, instruction.field] = instruction
        self.start_preprocessor(index, instruction.field)

    def start_preprocessor(self, index: str, field: str):
        if existing_task := self.running_tasks.get((index, field)):
            if not existing_task.done:
                return existing_task
        instruction = self.preprocessors[index, field]
        task = asyncio.create_task(run_processor_loop(index, instruction))
        self.running_tasks[index, instruction.field] = task

    def start_index_preprocessors(self, index: str):
        for ix, field in self.preprocessors:
            if ix == index:
                self.start_preprocessor(index, field)

    def stop_preprocessor(self, index: str, field: str):
        """Stop a preprocessor task"""
        try:
            if task := self.running_tasks.get((index, field)):
                task.cancel()
        except:
            logger.exception(f"Error on cancelling preprocessor {index}:{field}")

    def remove_preprocessor(self, index: str, field: str):
        """Stop this preprocessor remove them from the manager"""
        self.stop_preprocessor(index, field)
        del self.preprocessor_status[index, field]
        del self.preprocessors[index, field]

    def remove_index_preprocessors(self, index: str):
        """Stop all preprocessors on this index and remove them from the manager"""
        for ix, field in list(self.preprocessors.keys()):
            if index == ix:
                self.remove_preprocessor(ix, field)

    def get_status(self, index: str, field: str) -> PreprocessorStatus:
        status = self.preprocessor_status.get((index, field), "Unknown")
        task = self.running_tasks.get((index, field))
        if (not task) or task.done() and status == "Active":
            logger.warning(f"Preprocessor {index}.{field} is {status}, but has no running task: {task}")
            return "Unknown"
        return status


@cache
def get_manager():
    return PreprocessorManager()


def start_processors():
    logger.info("Starting preprocessing loops (if needed)")
    manager = get_manager()
    for index in amcat4.index.list_known_indices():
        try:
            instructions = list(amcat4.index.get_instructions(index.id))
        except NotFoundError:
            logger.warning(f"Index {index.id} does not exist!")
            continue
        for instruction in instructions:
            manager.add_preprocessor(index.id, instruction)


async def run_processor_loop(index, instruction: PreprocessingInstruction):
    """
    Main preprocessor loop.
    Calls process_documents to process a batch of documents, until 'done'
    """
    logger.info(f"Preprocessing START for {index}.{instruction.field}")
    get_manager().set_status(index, instruction.field, "Active")
    done = False
    while not done:
        try:
            done = await process_documents(index, instruction)
        except asyncio.CancelledError:
            logger.info(f"Preprocessing CANCEL for {index}.{instruction.field} cancelled")
            get_manager().set_status(index, instruction.field, "Stopped")
            raise
        except RateLimit:
            logger.info(f"Peprocessing RATELIMIT  for {index}.{instruction.field}")
            get_manager().set_status(index, instruction.field, "Paused")
            await asyncio.sleep(PAUSE_ON_RATE_LIMIT_SECONDS)
            get_manager().set_status(index, instruction.field, "Active")
        except Exception:
            logger.exception(f"Preprocessing ERROR for {index}.{instruction.field}")
            get_manager().set_status(index, instruction.field, "Error")
            return
    get_manager().set_status(index, instruction.field, "Done")
    logger.info(f"Preprocessing DONE for {index}.{instruction.field}")


async def process_documents(index: str, instruction: PreprocessingInstruction, size=100):
    """
    Process a batch of currently to-do documents in the index for this instruction.
    Return value indicates job completion:
    It returns True when it runs out of documents to do, or False if there might be more documents.
    """
    # Refresh index before getting new documents to make sure status updates are reflected
    amcat4.index.refresh_index(index)
    docs = list(get_todo(index, instruction, size=size))
    if not docs:
        return True
    logger.debug(f"Preprocessing for {index}.{instruction.field}: retrieved {len(docs)} docs to process")
    for doc in docs:
        await process_doc(index, instruction, doc)
    return False


def get_todo(index: str, instruction: PreprocessingInstruction, size=100):
    fields = [arg.field for arg in instruction.arguments if arg.field]
    q = dict(bool=dict(must_not=dict(exists=dict(field=instruction.field))))
    for doc in es().search(index=index, size=size, source_includes=fields, query=q)["hits"]["hits"]:
        yield {"_id": doc["_id"], **doc["_source"]}


def get_counts(index: str, field: str):
    agg = dict(status=dict(terms=dict(field=f"{field}.status")))

    res = es().search(index=index, size=0, aggs=agg)
    result = dict(total=res["hits"]["total"]["value"])
    for bucket in res["aggregations"]["status"]["buckets"]:
        result[bucket["key"]] = bucket["doc_count"]
    return result


async def process_doc(index: str, instruction: PreprocessingInstruction, doc: dict):
    # TODO catch errors and add to status field, rather than raising
    try:
        req = instruction.build_request(index, doc)
    except Exception as e:
        logger.exception(f"Error on preprocessing {index}.{instruction.field} doc {doc['_id']}")
        amcat4.index.update_document(index, doc["_id"], {instruction.field: dict(status="error", error=str(e))})
        return
    try:
        response = await AsyncClient().send(req)
        response.raise_for_status()
    except HTTPStatusError as e:
        if e.response.status_code == 503:
            raise RateLimit(e)
        logging.exception(f"Error on preprocessing {index}.{instruction.field} doc {doc['_id']}: {e.response.text}")
        body = dict(status="error", status_code=e.response.status_code, response=e.response.text)
        amcat4.index.update_document(index, doc["_id"], {instruction.field: body})
        return
    result = dict(instruction.parse_output(response.json()))
    result[instruction.field] = dict(status="done")
    amcat4.index.update_document(index, doc["_id"], result)



#!/usr/bin/env python

from distutils.core import setup

setup(
    name="amcat4",
    version="4.0.15",
    description="API for AmCAT4 Text Analysis",
    author="Wouter van Atteveldt",
    author_email="wouter@vanatteveldt.com",
    packages=["amcat4", "amcat4.api", "amcat4.preprocessing"],
    include_package_data=True,
    zip_safe=False,
    keywords=["API", "text"],
    classifiers=[
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Topic :: Text Processing",
    ],
    install_requires=[
        "fastapi[all]",
        "elasticsearch~=8.6",
        "python-multipart",
        "python-dotenv",
        "requests",
        "authlib",
        "pydantic[email]",
        "pydantic-settings",
        "typing_extensions",
        "uvicorn",
        "requests",
        "class_doc",
        "mypy",
        "minio",
        "jsonpath_ng",
    ],
    extras_require={
        "dev": [
            "pytest",
            "pytest-httpx",
            "mypy",
            "flake8",
            "responses",
            "pre-commit",
            "types-requests",
            "pytest-asyncio",
            "pytest-minio-mock @ git+ssh://git@github.com/vanatteveldt/pytest-minio-mock.git",
        ]
    },
    entry_points={"console_scripts": ["amcat4 = amcat4.__main__:main"]},
)



import copy
from multiprocessing import Value
from typing import Any, Iterable, List, Optional, Tuple

import httpx
from pydantic import BaseModel

from amcat4 import multimedia
from amcat4.fields import get_fields
from amcat4.preprocessing.task import get_task


class PreprocessingArgument(BaseModel):
    name: str
    field: Optional[str] = None
    value: Optional[str | int | bool | float | List[str] | List[int] | List[float]] = None
    secret: Optional[bool] = False


class PreprocessingOutput(BaseModel):
    name: str
    field: str


class PreprocessingInstruction(BaseModel):
    field: str
    task: str
    endpoint: str
    arguments: List[PreprocessingArgument]
    outputs: List[PreprocessingOutput]

    def build_request(self, index, doc) -> httpx.Request:
        # TODO: validate that instruction is valid for task!
        fields = get_fields(index)
        task = get_task(self.task)
        if task.request.body == "json":
            if not task.request.template:
                raise ValueError(f"Task {task.name} has json body but not template")
            body = copy.deepcopy(task.request.template)
        elif task.request.body == "binary":
            body = None
        else:
            raise NotImplementedError()
        headers = {}
        for argument in self.arguments:
            param = task.get_parameter(argument.name)
            if param.use_field == "yes":
                if not argument.field:
                    raise ValueError("Field not given for field param")
                value = doc.get(argument.field)
                if task.request.body == "binary" and fields[argument.field].type in ["image"]:
                    value = multimedia.get_multimedia_object(index, value)
            else:
                value = argument.value
            if param.header:
                if not param.path:
                    raise ValueError("Path required for header params")
                if ":" in param.path:
                    path, prefix = param.path.split(":", 1)
                    prefix = f"{prefix} "
                else:
                    path, prefix = param.path, ""
                headers[path] = f"{prefix}{value}"
            else:
                if task.request.body == "json":
                    if not param.path:
                        raise ValueError("Path required for json body params")
                    param.parsed.update(body, value)
                elif task.request.body == "binary":
                    if param.path:
                        raise ValueError("Path not allowed for binary body")
                    if body:
                        raise ValueError("Multiple values for body")
                    if type(value) != bytes:
                        raise ValueError("Binary request requires multimedia object")
                    body = value
        if task.request.body == "json":
            return httpx.Request("POST", self.endpoint, json=body, headers=headers)
        else:
            return httpx.Request("POST", self.endpoint, content=body, headers=headers)

    def parse_output(self, output) -> Iterable[Tuple[str, Any]]:
        task = get_task(self.task)
        for arg in self.outputs:
            o = task.get_output(arg.name)
            yield arg.field, o.parsed.find(output)[0].value



"""
Aggregate queries
"""

import copy
from datetime import datetime
from typing import Any, Mapping, Iterable, Union, Tuple, Sequence, List, Dict

from amcat4.date_mappings import interval_mapping
from amcat4.elastic import es
from amcat4.fields import get_fields
from amcat4.query import build_body
from amcat4.models import Field, FilterSpec


def _combine_mappings(mappings):
    result = {}
    for mapping in mappings:
        if mapping:
            result.update(mapping)
    return result


class Axis:
    """
    Class that specifies an aggregation axis
    """

    def __init__(self, field: str, interval: str | None = None, name: str | None = None, field_type: str | None = None):
        self.field = field
        self.interval = interval
        self.ftype = field_type
        if name:
            self.name = name
        elif interval:
            self.name = f"{field}_{interval}"
        else:
            self.name = field

    def __repr__(self):
        return f"<Axis field={self.field} ftype={self.ftype}>"

    def query(self):
        if not self.ftype:
            raise ValueError("Please set index before using axis")
        if self.interval:
            if self.ftype == "date":
                if m := interval_mapping(self.interval):
                    return {self.name: {"terms": {"field": m.fieldname(self.field)}}}
                return {self.name: {"date_histogram": {"field": self.field, "calendar_interval": self.interval}}}
            else:
                return {self.name: {"histogram": {"field": self.field, "interval": self.interval}}}
        else:
            return {self.name: {"terms": {"field": self.field, "order": "desc"}}}

    def get_value(self, values):
        value = values[self.name]
        if m := interval_mapping(self.interval):
            value = m.postprocess(value)
        elif self.ftype == "date":
            value = datetime.utcfromtimestamp(value / 1000.0)
            if self.interval in {"year", "month", "week", "day"}:
                value = value.date()
        return value

    def asdict(self):
        return {"name": self.name, "field": self.field, "type": self.ftype, "interval": self.interval}

    def runtime_mappings(self):
        if m := interval_mapping(self.interval):
            return m.mapping(self.field)


class Aggregation:
    """
    Specification of a single aggregation, that is, field and aggregation function
    """

    def __init__(self, field: str, function: str, name: str | None = None, ftype: str | None = None):
        self.field = field
        self.function = function
        self.name = name or f"{function}_{field}"
        self.ftype = ftype

    def dsl_item(self):
        return self.name, {self.function: {"field": self.field}}

    def get_value(self, bucket: dict):
        result = bucket[self.name]["value"]
        if result and self.ftype == "date":
            result = datetime.utcfromtimestamp(result / 1000.0)
        return result

    def asdict(self):
        return {"field": self.field, "type": self.ftype, "function": self.function, "name": self.name}


def aggregation_dsl(aggregations: Iterable[Aggregation]) -> dict:
    """Get the aggregation DSL dict for a list of aggregations"""
    return dict(a.dsl_item() for a in aggregations)


class AggregateResult:
    def __init__(
        self,
        axes: Sequence[Axis],
        aggregations: List[Aggregation],
        data: List[tuple],
        count_column: str = "n",
        after: dict | None = None,
    ):
        self.axes = axes
        self.data = data
        self.aggregations = aggregations
        self.count_column = count_column
        self.after = after

    def as_dicts(self) -> Iterable[dict]:
        """Return the results as a sequence of {axis1, ..., n} dicts"""
        keys = tuple(ax.name for ax in self.axes) + (self.count_column,)
        if self.aggregations:
            keys += tuple(a.name for a in self.aggregations)
        for row in self.data:
            yield dict(zip(keys, row))


def _bare_aggregate(index: str | list[str], queries, filters, aggregations: Sequence[Aggregation]) -> Tuple[int, dict]:
    """
    Aggregate without sources/group_by.
    Returns a tuple of doc count and aggregegations (doc_count, {metric: value})
    """
    body = build_body(queries=queries, filters=filters) if filters or queries else {}
    index = index if isinstance(index, str) else ",".join(index)
    aresult = es().search(index=index, size=0, aggregations=aggregation_dsl(aggregations), **body)
    cresult = es().count(index=index, **body)
    return cresult["count"], aresult["aggregations"]


def _elastic_aggregate(
    index: str | list[str],
    sources,
    axes,
    queries,
    filters,
    aggregations: list[Aggregation],
    runtime_mappings: dict[str, Mapping] | None = None,
    after_key=None,
) -> Tuple[list, dict | None]:
    """
    Recursively get all buckets from a composite query.
    Yields 'buckets' consisting of {key: {axis: value}, doc_count: <number>}
    """
    # [WvA] Not sure if we should get all results ourselves or expose the 'after' pagination.
    #       This might get us in trouble if someone e.g. aggregates on url or day for a large corpus
    after = {"after": after_key} if after_key is not None and len(after_key) > 0 else {}
    aggr: Dict[str, Dict[str, dict]] = {"aggs": {"composite": dict(sources=sources, **after)}}
    if aggregations:
        aggr["aggs"]["aggregations"] = aggregation_dsl(aggregations)
    kargs = {}

    if filters or queries:
        q = build_body(queries=queries, filters=filters)
        kargs["query"] = q["query"]

    result = es().search(
        index=index if isinstance(index, str) else ",".join(index),
        size=0,
        aggregations=aggr,
        runtime_mappings=runtime_mappings,
        **kargs,
    )
    if failure := result.get("_shards", {}).get("failures"):
        raise Exception(f"Error on running aggregate search: {failure}")

    buckets = result["aggregations"]["aggs"]["buckets"]
    after_key = result["aggregations"]["aggs"].get("after_key")

    rows = []
    for bucket in buckets:
        row = tuple(axis.get_value(bucket["key"]) for axis in axes)
        row += (bucket["doc_count"],)
        if aggregations:
            row += tuple(a.get_value(bucket) for a in aggregations)
        rows.append(row)

    return rows, after_key


def _aggregate_results(
    index: Union[str, List[str]],
    axes: List[Axis],
    queries: dict[str, str] | None,
    filters: dict[str, FilterSpec] | None,
    aggregations: List[Aggregation],
    after: dict[str, Any] | None = None,
):

    if not axes or len(axes) == 0:
        # Path 1
        # No axes, so return aggregations (or total count) only
        if aggregations:
            count, results = _bare_aggregate(index, queries, filters, aggregations)
            rows = [(count,) + tuple(a.get_value(results) for a in aggregations)]
        else:
            result = es().count(
                index=index if isinstance(index, str) else ",".join(index), **build_body(queries=queries, filters=filters)
            )
            rows = [(result["count"],)]
        yield rows, None

    elif any(ax.field == "_query" for ax in axes):

        # Path 2
        # We cannot run the aggregation for multiple queries at once, so we loop over queries
        # and recursively call _aggregate_results with one query at a time (which then uses path 3).
        if queries is None:
            raise ValueError("Queries must be specified when aggregating by query")
        # Strip off _query axis and run separate aggregation for each query
        i = [ax.field for ax in axes].index("_query")
        _axes = axes[:i] + axes[(i + 1) :]

        query_items = list(queries.items())
        for label, query in query_items:
            last_query = label == query_items[-1][0]

            if after is not None and "_query" in after:
                # after is a dict with the aggregation values from which to continue
                # pagination. Since we loop over queries, we add the _query value.
                # Then after continuing from the right query, we remove this _query
                # key so that the after dict is as elastic expects it
                if after.get("_query") != label:
                    continue
                after.pop("_query", None)

            for rows, after_buckets in _aggregate_results(index, _axes, {label: query}, filters, aggregations, after=after):
                after_buckets = copy.deepcopy(after_buckets)

                # insert label into the right position on the result tuple
                rows = [result_tuple[:i] + (label,) + result_tuple[i:] for result_tuple in rows]

                if after_buckets is None:
                    # if there are no buckets left for this query, we check if this is the last query.
                    # If not, we need to return the _query value to ensure pagination continues from this query
                    if not last_query:
                        after_buckets = {"_query": label}
                else:
                    # if there are buckets left, we add the _query value to ensure pagination continues from this query
                    after_buckets["_query"] = label
                yield rows, after_buckets

            # after only applies to the first query
            after = None

    else:
        # Path 3
        # Run an aggregation with one or more axes. If after is not None, we continue from there.
        sources = [axis.query() for axis in axes]
        runtime_mappings = _combine_mappings(axis.runtime_mappings() for axis in axes)

        rows, after = _elastic_aggregate(index, sources, axes, queries, filters, aggregations, runtime_mappings, after)
        yield rows, after

        if after is not None:
            for rows, after in _aggregate_results(index, axes, queries, filters, aggregations, after):
                yield rows, after


def query_aggregate(
    index: str | list[str],
    axes: list[Axis] | None = None,
    aggregations: list[Aggregation] | None = None,
    *,
    queries: dict[str, str] | None = None,
    filters: dict[str, FilterSpec] | None = None,
    after: dict[str, Any] | None = None,
) -> AggregateResult:
    """
    Conduct an aggregate query.
    Note that interval queries also yield zero counts for intervening keys without value,
    but only if that is the last axis. [WvA] Not sure if this is desired

    :param index: The name of the elasticsearch index
    :param axes: Aggregation axes
    :param aggregations: Aggregation fields
    :param queries: Optional query string
    :param filters: if not None, a dict of filters: {field: {'value': value}} or
                    {field: {'range': {'gte/gt/lte/lt': value, 'gte/gt/..': value, ..}}
    :return: a pair of (Axis, results), where results is a sequence of tuples
    """
    if axes and len([x.field == "_query" for x in axes[1:]]) > 1:
        raise ValueError("Only one aggregation axis may be by query")

    all_fields: dict[str, Field] = dict()
    indices = index if isinstance(index, list) else [index]
    for index in indices:
        index_fields = get_fields(index)
        for field_name, field in index_fields.items():
            if field_name not in all_fields:
                all_fields[field_name] = field
            else:
                if field.type != all_fields[field_name].type:
                    raise ValueError(f"Type of {field_name} is not the same in all indices")
        all_fields.update(get_fields(index))

    if not axes:
        axes = []
    for axis in axes:
        axis.ftype = "_query" if axis.field == "_query" else all_fields[axis.field].type
    if not aggregations:
        aggregations = []
    for aggregation in aggregations:
        aggregation.ftype = all_fields[aggregation.field].type

    # We get the rows in sets of queries * buckets, and if there are queries or buckets left,
    # the last_after value serves as a pagination cursor. Once we have > [stop_after] rows,
    # we return the data and the last_after cursor. If the user needs to collect the rest,
    # they need to paginate
    stop_after = 1000
    gen = _aggregate_results(indices, axes, queries, filters, aggregations, after)
    data = list()
    last_after = None
    for rows, after in gen:
        data += rows
        last_after = after
        if len(data) > stop_after:
            gen.close()
    return AggregateResult(axes, aggregations, data, count_column="n", after=last_after)
